# -*- coding: utf-8 -*-
#
# Testing module for ACME's dask components
#

# Builtin/3rd party package imports
import pytest
import getpass
import numpy as np

# Import main actors here
from acme.dask_helpers import customIOError
from acme import cluster_cleanup, esi_cluster_setup, slurm_cluster_setup, local_cluster_setup
from conftest import useSLURM, onESI, defaultQ

def test_cluster_setup():

    # Tests which should work on any SLURM cluster
    if useSLURM:

        # Ensure invalid partition/memory specifications are caught
        with pytest.raises(ValueError):
            slurm_cluster_setup(partition="invalid")
        cluster_cleanup()
        with pytest.raises(TypeError):
            slurm_cluster_setup(partition=defaultQ, mem_per_worker=(3))
        cluster_cleanup()
        with pytest.raises(ValueError):
            slurm_cluster_setup(partition=defaultQ, mem_per_worker="invalidGB")
        cluster_cleanup()
        with pytest.raises(ValueError):
            slurm_cluster_setup(partition=defaultQ, mem_per_worker="-110MB")
        cluster_cleanup()

        # Ensure additional sbatch parameters are processed correctly
        with pytest.raises(TypeError):
            slurm_cluster_setup(partition=defaultQ, job_extra="--output")
        cluster_cleanup()
        with pytest.raises(ValueError):
            slurm_cluster_setup(partition=defaultQ, job_extra=["invalid"])
        cluster_cleanup()
        client = slurm_cluster_setup(partition=defaultQ,
                                     job_extra=["--job-name='averycustomjobname'"],
                                     interactive=False)
        assert 'averycustomjobname' in client.cluster.job_header
        cluster_cleanup()

        # Ensure output directory specification is parsed for correctness
        with pytest.raises(ValueError):
            slurm_cluster_setup(partition=defaultQ, job_extra=["--output=/path/to/nowhere"])
        cluster_cleanup()
        slurmOut = "/tmp/{}".format(getpass.getuser())
        client = slurm_cluster_setup(partition=defaultQ,
                                     n_workers=1,
                                     timeout=120,
                                     job_extra=["--output={}".format(slurmOut), "--job-name='averycustomjobname'"],
                                     interactive=False)
        assert "--output={}".format(slurmOut) in client.cluster.job_header
        assert 'averycustomjobname' in client.cluster.job_header
        cluster_cleanup(client)

        # Tests specific to the ESI HPC cluster
        if onESI:

            # Over-allocation of memory should default to partition max
            # (this should work on all clusters but we don't know partition
            # names, QoS rules etc.)
            client = esi_cluster_setup(partition="8GBDEV",
                                       timeout=120,
                                       n_workers=1,
                                       mem_per_worker="9000MB",
                                       interactive=False)
            memory = np.unique([w["memory_limit"] for w in client.cluster.scheduler_info["workers"].values()])
            assert memory.size == 1
            assert np.round(memory / 1000**3)[0] == 8

            # Invoking `esi_cluster_setup` with existing client must not start a new one
            clnt = esi_cluster_setup(partition="16GBXS", n_workers=2, interactive=False)
            assert clnt == client
            cluster_cleanup(client)

            # Specify CPU count manually
            client = esi_cluster_setup(partition="16GBDEV",
                                       timeout=120,
                                       n_workers=1,
                                       n_cores=1,
                                       interactive=False)
            assert [w["nthreads"] for w in client.cluster.scheduler_info["workers"].values()][0] == 1
            cluster_cleanup(client)

    else:

        # Trigger an exception by invoking `slurm_cluster_setup`` on non-SLURM node
        with pytest.raises(customIOError) as err:
            slurm_cluster_setup()
            assert "Cannot access SLURM queuing system" in str(err)


def test_local_setup():

    # Allocate local distributed computing client w/custom settings
    client = local_cluster_setup(n_workers=2, mem_per_worker="2GB", interactive=False)
    wMem = [w["memory_limit"] for w in client.cluster.scheduler_info["workers"].values()]
    assert len(wMem) == 2
    memory = np.unique(wMem)
    assert memory.size == 1
    assert np.round(memory / 1000**3)[0] == 2
    cluster_cleanup()

    # Allocate local distributed computing client w/default settings
    client = local_cluster_setup(interactive=False)
    assert len(client.cluster.scheduler_info["workers"].keys()) > 1
    cluster_cleanup()


def test_backcompat_cluster():

    if useSLURM:
        client = slurm_cluster_setup(partition=defaultQ,
                                     timeout=120,
                                     n_jobs=1,
                                     workers_per_job=1,
                                     mem_per_job="1GB",
                                     n_jobs_startup=1,
                                     interactive=False)
        assert len(client.cluster.workers) == 1
        memory = [w["memory_limit"] for w in client.cluster.scheduler_info["workers"].values()]
        assert round(memory[0] / 1000**3) == 1
        threads = [w["nthreads"] for w in client.cluster.scheduler_info["workers"].values()]
        assert threads[0] == 1
        cluster_cleanup(client)



# -*- coding: utf-8 -*-
#
# Central pytest configuration
#

# Builtin/3rd party package imports
import sys
import pytest

# Import main actors here
from acme.shared import is_slurm_node, is_esi_node

# Construct decorators for skipping certain tests
skip_if_not_linux = pytest.mark.skipif(sys.platform != "linux", reason="Only works in Linux")

# Perform SLURM-specific tests only on cluster nodes
useSLURM = is_slurm_node()

# Perform ESI-specific tests only the ESI HPC cluster
onESI = is_esi_node()

# Use a default partition if running on the ESI cluster
if onESI:
    defaultQ ="8GBXS"
else:
    defaultQ ="auto"



# -*- coding: utf-8 -*-
#
# Testing module for ACME's `ParallelMap` interface
#

# Builtin/3rd party package imports
import os
import sys
import platform
import pickle
import shutil
import inspect
import subprocess
import getpass
import time
import itertools
import logging
import h5py
import pytest
import signal as sys_signal
import numpy as np
import dask.distributed as dd
from glob import glob
from scipy import signal

# Import main actors here
from acme import ParallelMap, cluster_cleanup, esi_cluster_setup
from acme.dask_helpers import customIOError
from conftest import skip_if_not_linux, useSLURM, onESI, defaultQ

# Functions that act as stand-ins for user-funcs
def simple_func(x, y, z=3):
    return (x + y) * z

def medium_func(x, y, z=3, w=np.ones((3, 3))):
    return (sum(x) + y) * z * w.max()

def hard_func(x, y, z=3, w=np.zeros((3, 1)), **kwargs):
    return sum(x) + y,  z * w

def lowpass_simple(h5name, channel_no):
    with h5py.File(h5name, "r") as h5f:
        channel = h5f["data"][:, channel_no]
        b = h5f["data"].attrs["b"]
        a = h5f["data"].attrs["a"]
    res = signal.filtfilt(b, a, channel, padlen=200)
    return res

def lowpass_medium(h5name, channel_no):
    with h5py.File(h5name, "r") as h5f:
        channel = h5f["data"][:, channel_no]
        b = h5f["data"].attrs["b"]
        a = h5f["data"].attrs["a"]
    res = signal.filtfilt(b, a, channel, padlen=200)
    return res, channel_no, b, a

def lowpass_hard(arr_like, b, a, res_dir, res_base="lowpass_hard_", dset_name="custom_dset_name", padlen=200, taskID=None):
    channel = arr_like[:, taskID]
    res = signal.filtfilt(b, a, channel, padlen=padlen)
    h5name = os.path.join(res_dir, res_base +"{}.h5".format(taskID))
    with h5py.File(h5name, "w") as h5f:
        h5f.create_dataset(dset_name, data=res)
    return

def pickle_func(arr, b, a, channel_no, sabotage_hdf5=False):
    res = signal.filtfilt(b, a, arr[:, channel_no], padlen=200)
    if sabotage_hdf5:
        if channel_no % 2 == 0:
            return {"b" : b}
    return res

def memtest_func(x, y, z=3, arrsize=2, sleeper=300):
    fSize = np.dtype("float").itemsize
    time.sleep(2)
    arr = np.ones((int(arrsize * 1024**3 / fSize), ))   # `arrsize` denotes array size in GB
    time.sleep(sleeper)
    return (x + y) * z * arr.max()


# Main testing class
class TestParallelMap():

    # Construct linear combination of low- and high-frequency sine waves
    # and use an IIR filter to reconstruct the low-frequency component
    nChannels = 32
    nTrials = 8
    fData = 2
    fNoise = 64
    fs = 1000
    t = np.linspace(-1, 1, fs)
    orig = np.sin(2 * np.pi * fData * t)
    sig = orig + np.sin(2 * np.pi * fNoise * t)
    cutoff = 50
    b, a = signal.butter(8, 2 * cutoff / fs)

    # Blow up the signal to have "channels" and "trials": even/odd channels have
    # opposing periodicity; do the same to the low-freq component
    sig = np.repeat(sig.reshape(-1, 1), axis=1, repeats=nChannels)
    sig[:, ::2] *= -1
    sig = np.tile(sig, (nTrials, 1))
    orig = np.repeat(orig.reshape(-1, 1), axis=1, repeats=nChannels)
    orig[:, ::2] *= -1
    orig = np.tile(orig, (nTrials, 1))

    # Error tolerance for low-pass filtered results
    tol = 1e-3

    # Helper method for allocating data containers
    def _prep_data(self, tmpName):

        # Create tmp directory and create data-containers
        tempDir = os.path.join(os.path.abspath(os.path.expanduser("~")), tmpName)
        if useSLURM:
            tempDir = "/cs/home/{}/{}".format(getpass.getuser(), tmpName)
        os.makedirs(tempDir, exist_ok=True)
        sigName = os.path.join(tempDir, "sigdata.h5")
        origName = os.path.join(tempDir, "origdata.h5")
        with h5py.File(sigName, "w") as sigFile:
            dset = sigFile.create_dataset("data", data=self.sig)
            dset.attrs["b"] = self.b
            dset.attrs["a"] = self.a
        with h5py.File(origName, "w") as origFile:
            origFile.create_dataset("data", data=self.orig)

        return tempDir, sigName

    # Test setup of `ParallelMap` w/different functions args/kwargs
    def test_init(self, testclient=None):

        # Collected auto-generated output directories in list for later cleanup
        outDirs = []

        # Basic functionality w/simplest conceivable user-func
        pmap = ParallelMap(simple_func, [2, 4, 6, 8], 4, partition=defaultQ, setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)
        pmap = ParallelMap(simple_func, [2, 4, 6, 8], y=4, partition=defaultQ, setup_interactive=False)  # pos arg referenced via kwarg, cfg #2
        outDirs.append(pmap.daemon.out_dir)
        pmap = ParallelMap(simple_func, 0, 4, z=[3, 4, 5, 6], partition=defaultQ, setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)
        pmap = ParallelMap(simple_func, [2, 4, 6, 8], [2, 2], n_inputs=2, partition=defaultQ, setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)

        # User func has `np.ndarray` as keyword
        pmap = ParallelMap(medium_func, [2, 4, 6, 8], y=[2, 2], n_inputs=2, partition=defaultQ, setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)
        pmap = ParallelMap(medium_func, None, None, w=[np.ones((3, 3)), 2 * np.ones((3,3))], partition=defaultQ, setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)
        pmap = ParallelMap(medium_func, None, None, z=np.zeros((3,)), partition=defaultQ, setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)
        pmap = ParallelMap(medium_func, None, None, z=np.zeros((3, 1)), partition=defaultQ, setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)

        # Lots of ways for this to go wrong...
        pmap = ParallelMap(hard_func, [2, 4, 6, 8], 2, w=np.ones((3,)), partition=defaultQ, setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)
        pmap = ParallelMap(hard_func, [2, 4, 6, 8], y=22, w=np.ones((7, 1)), partition=defaultQ, setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)
        pmap = ParallelMap(hard_func, np.ones((3,)), 1, w=np.ones((7, 1)), partition=defaultQ, setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)
        pmap = ParallelMap(hard_func, [2, 4, 6, 8], [2, 2], z=np.array([1, 2]), w=np.ones((8, 1)), n_inputs=2, partition=defaultQ, setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)
        pmap = ParallelMap(hard_func, [2, 4, 6, 8], [2, 2], w=np.ones((8, 1)), n_inputs=4, partition=defaultQ, setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)

        # Ensure erroneous/ambiguous setups trigger the appropriate errors:
        # not enough positional args
        with pytest.raises(ValueError) as valerr:
            ParallelMap(simple_func, 4, setup_interactive=False)
            assert "simple_func expects 2 positional arguments ('x', 'y'), found 1" in str(valerr.value)
        # invalid kwargs
        with pytest.raises(ValueError) as valerr:
            ParallelMap(simple_func, 4, 4, z=3, w=4, setup_interactive=False)
            assert "simple_func accepts at maximum 1 keyword arguments ('z'), found 2" in str(valerr.value)
        # ill-posed parallelization: two candidate lists for input distribution
        with pytest.raises(ValueError) as valerr:
            ParallelMap(simple_func, [2, 4, 6, 8], [2, 2], setup_interactive=False)
            assert "automatic input distribution failed: found 2 objects containing 2 to 4 elements" in str(valerr.value)
        # ill-posed parallelization: two candidate lists for input distribution (`x` and `w`)
        with pytest.raises(ValueError) as valerr:
            ParallelMap(medium_func, [1, 2, 3], None, w=[np.ones((3,3)), 2 * np.ones((3,3))], setup_interactive=False)
            assert "automatic input distribution failed: found 2 objects containing 2 to 3 elements." in str(valerr.value)
        # invalid input spec
        with pytest.raises(ValueError) as valerr:
            ParallelMap(simple_func, [2, 4, 6, 8], [2, 2], n_inputs=3, setup_interactive=False)
            assert "No object has required length of 3 matching `n_inputs`" in str(valerr.value)
        # invalid input spec: `w` expects a NumPy array, thus it is not considered for input distribution
        with pytest.raises(ValueError) as valerr:
            ParallelMap(hard_func, [2, 4, 6, 8], [2, 2], w=np.ones((8, 1)), n_inputs=8, setup_interactive=False)
            assert "No object has required length of 8 matching `n_inputs`" in str(valerr.value)

        # Clean up testing folder and any running clients
        if testclient is None:
            cluster_cleanup()
        for folder in outDirs:
            shutil.rmtree(folder, ignore_errors=True)

        return testclient

    # Functionality tests: perform channel-concurrent low-pass filtering
    def test_simple_filter(self, testclient=None):

        # Prepare data containers
        tempDir, sigName = self._prep_data("acme_tmp")

        # Collect auto-generated output directories in list for later cleanup
        outDirs = []

        # Parallelize across channels, write results to disk
        with ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            resOnDisk = pmap.compute()
        outDirs.append(pmap.out_dir)

        # Ensure each compute run generated a dedicated HDF5 file
        assert len(pmap.kwargv["outFile"]) == pmap.n_calls

        # Query auto-generated output directory
        outDirContents = glob(os.path.join(pmap.out_dir, "*"))
        payloadDir = pmap.results_container.replace(".h5", "_payload")
        assert pmap.results_container in outDirContents
        assert payloadDir in outDirContents
        resFiles = glob(os.path.join(payloadDir, "*.h5"))
        assert len(resFiles) == pmap.n_calls
        assert all(fle in resFiles for fle in resOnDisk)
        assert all(os.path.isfile(fle) for fle in resOnDisk)

        # Compare computed single-channel results to expected low-freq signal
        # and ensure collection container was assembled correctly
        with h5py.File(pmap.results_container, "r") as h5col:
            dset = "comp_{}/result_0"
            for chNo, h5name in enumerate(resOnDisk):
                with h5py.File(h5name, "r") as h5f:
                    assert np.mean(np.abs(h5f["result_0"][()] - self.orig[:, chNo])) < self.tol
                    assert np.array_equal(h5col[dset.format(chNo)][()], h5f["result_0"][()])

        # Remember results for later use
        colRes = str(pmap.results_container)
        colResPayload = str(payloadDir)

        # Same with `single_file`
        with ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         partition=defaultQ,
                         setup_interactive=False,
                         single_file=True) as pmap:
            singleResOnDisk = pmap.compute()
        outDirs.append(pmap.out_dir)

        # Ensure only one file was generated
        assert len(singleResOnDisk) == 1
        outDirContents = glob(os.path.join(pmap.out_dir, "*"))
        assert outDirContents == singleResOnDisk
        assert pmap.results_container in outDirContents[0]
        assert os.path.isfile(singleResOnDisk[0])

        # Compare results to container computed above
        with h5py.File(colRes, "r") as h5col:
            with h5py.File(pmap.results_container, "r") as h5single:
                dset = "comp_{}/result_0"
                for chNo in range(self.nChannels):
                    assert np.array_equal(h5single[dset.format(chNo)][()], h5col[dset.format(chNo)][()])

        # Now use non-standard output directory
        outDir = os.path.join(tempDir, "somewhere")
        with ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         output_dir=outDir,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            resOnDisk = pmap.compute()

        # Query specified custom output directory
        assert pmap.out_dir == outDir
        outDirContents = glob(os.path.join(pmap.out_dir, "*"))
        payloadDir = pmap.results_container.replace(".h5", "_payload")
        assert pmap.results_container in outDirContents
        assert payloadDir in outDirContents
        resFiles = glob(os.path.join(payloadDir, "*.h5"))
        assert len(resFiles) == pmap.n_calls
        assert all(fle in resFiles for fle in resOnDisk)
        assert all(os.path.isfile(fle) for fle in resOnDisk)

        # A little overly paranoid, but compare results still...
        with h5py.File(colRes, "r") as h5col:
            with h5py.File(pmap.results_container, "r") as h5comp:
                dset = "comp_{}/result_0"
                for chNo in range(self.nChannels):
                    assert np.array_equal(h5comp[dset.format(chNo)][()], h5col[dset.format(chNo)][()])

        # Finally collect results in memory: ensure nothing freaky happens
        with ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         write_worker_results=False,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            resInMem = pmap.compute()

        # Be double-paranoid: ensure on-disk and in-memory results match up
        with h5py.File(colRes, "r") as h5col:
            dset = "comp_{}/result_0"
            for chNo in range(self.nChannels):
                assert np.array_equal(h5col[dset.format(chNo)][()], resInMem[chNo])

        # Comparisons are over, now remove payload and ensure container is broken
        shutil.rmtree(colResPayload)
        with pytest.raises(KeyError) as keyerr:
            with h5py.File(colRes, "r") as h5col:
                chNo = np.random.choice(self.nChannels, size=1)[0]
                h5col["comp_{}".format(chNo)]["result_0"]
            assert "unable to open external file" in str(keyerr.value)

        # Ensure `output_dir` is properly ignored if `write_worker_results` is `False`
        pmap = ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         output_dir=tempDir,
                         write_worker_results=False,
                         partition=defaultQ,
                         setup_interactive=False)
        assert pmap.daemon.out_dir is None
        assert pmap.daemon.collect_results is True

        # Simulate user-defined results-directory not auto-populated by ACME
        tempDir2 = os.path.join(os.path.abspath(os.path.expanduser("~")), "acme_tmp_lowpass_hard")
        if useSLURM:
            tempDir2 = "/cs/home/{}/acme_tmp_lowpass_hard".format(getpass.getuser())
        os.makedirs(tempDir2, exist_ok=True)

        # Same task, different function: simulate user-defined saving scheme and "weird" inputs
        sigData = h5py.File(sigName, "r")["data"]
        res_base = "lowpass_hard_"
        dset_name = "custom_dset_name"
        with ParallelMap(lowpass_hard,
                         sigData,
                         self.b,
                         self.a,
                         res_dir=tempDir2,
                         res_base=res_base,
                         dset_name=dset_name,
                         padlen=[200] * self.nChannels,
                         n_inputs=self.nChannels,
                         write_worker_results=False,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            pmap.compute()
        resFiles = glob(os.path.join(tempDir2, res_base + "*"))
        assert len(resFiles) == pmap.n_calls

        # Compare computed single-channel results to expected low-freq signal
        for chNo in range(self.nChannels):
            h5name = res_base + "{}.h5".format(chNo)
            with h5py.File(os.path.join(tempDir2, h5name), "r") as h5f:
                assert np.mean(np.abs(h5f[dset_name][()] - self.orig[:, chNo])) < self.tol

        # Ensure log-file generation produces a non-empty log-file at the expected location
        # Bonus: leave computing client alive and vet default SLURM settings
        if testclient is None:
            cluster_cleanup(pmap.client)
        for handler in pmap.log.handlers:
            if isinstance(handler, logging.FileHandler):
                pmap.log.handlers.remove(handler)
        with ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         logfile=True,
                         stop_client=False,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            pmap.compute()
        outDirs.append(pmap.out_dir)
        logFileList = [handler.baseFilename for handler in pmap.log.handlers if isinstance(handler, logging.FileHandler)]
        assert len(logFileList) == 1
        logFile = logFileList[0]
        assert os.path.dirname(os.path.realpath(__file__)) in logFile
        with open(logFile, "r") as fl:
            assert len(fl.readlines()) > 1

        # Ensure client has not been killed; perform post-hoc check of default SLURM settings
        assert dd.get_client()
        client = dd.get_client()
        if useSLURM is True and testclient is None:
            assert pmap.n_calls == pmap.n_workers
            assert len(client.cluster.workers) == pmap.n_workers
            partition = client.cluster.job_header.split("-p ")[1].split("\n")[0]
            assert "8GB" in partition
            memory = np.unique([w["memory_limit"] for w in client.cluster.scheduler_info["workers"].values()])
            assert memory.size == 1
            assert round(memory[0] / 1000**3) == [int(s) for s in partition if s.isdigit()][0]

        # Wait a sec (literally) for dask to collect its bearings (after the
        # `get_client` above) before proceeding
        time.sleep(1.0)

        # Same, but use custom log-file
        customLog = os.path.join(tempDir, "acme_log.txt")
        with ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         logfile=customLog,
                         verbose=True,
                         stop_client=testclient is None,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            pmap.compute()
        outDirs.append(pmap.out_dir)
        assert os.path.isfile(customLog)
        with open(customLog, "r") as fl:
            assert len(fl.readlines()) > 1

        # Ensure only single log file `customLog` is used
        assert len([h for h in pmap.log.handlers if isinstance(h, logging.FileHandler)]) == 1

        # Ensure client has been stopped
        if testclient is None:
            with pytest.raises(ValueError):
                dd.get_client()

        # Wait a sec (literally) to give dask enough time to close the client
        time.sleep(1.0)

        # Underbook SLURM (more calls than workers)
        partition = "8GBXS"
        n_workers = int(self.nChannels / 2)
        mem_per_worker = "2GB"
        with ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         partition=partition,
                         n_workers=n_workers,
                         mem_per_worker=mem_per_worker,
                         stop_client=False,
                         setup_interactive=False) as pmap:
            pmap.compute()
        outDirs.append(pmap.out_dir)

        # Post-hoc check of client to ensure custom settings were respected
        client = pmap.client
        assert pmap.n_calls == self.nChannels
        if useSLURM is True and testclient is None:
            assert pmap.n_workers == n_workers
            assert len(client.cluster.workers) == pmap.n_workers
            actualPartition = client.cluster.job_header.split("-p ")[1].split("\n")[0]
            assert actualPartition == partition
            memory = np.unique([w["memory_limit"] for w in client.cluster.scheduler_info["workers"].values()])
            assert memory.size == 1
            assert round(memory[0] / 1000**3) == int(mem_per_worker.replace("GB", ""))

        # Let `cluster_cleanup` murder the custom setup and ensure it did its job
        if testclient is None:
            cluster_cleanup(pmap.client)
            with pytest.raises(ValueError):
                dd.get_client()

        # Overbook SLURM (more workers than calls)
        partition = "8GBXS"
        n_workers = self.nChannels + 2
        mem_per_worker = "3000MB"
        with ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         partition=partition,
                         n_workers=n_workers,
                         mem_per_worker=mem_per_worker,
                         stop_client=False,
                         setup_interactive=False) as pmap:
            pmap.compute()
        outDirs.append(pmap.out_dir)

        # Post-hoc check of client to ensure custom settings were respected
        client = pmap.client
        assert pmap.n_calls == self.nChannels
        if useSLURM and testclient is None:
            assert pmap.n_workers == n_workers
            assert len(client.cluster.workers) == pmap.n_workers
            actualPartition = client.cluster.job_header.split("-p ")[1].split("\n")[0]
            assert actualPartition == partition
            memory = np.unique([w["memory_limit"] for w in client.cluster.scheduler_info["workers"].values()])
            assert memory.size == 1
            assert round(memory[0] / 1000**3) * 1000 == int(mem_per_worker.replace("MB", ""))
        if testclient is None:
            cluster_cleanup(pmap.client)

        # Close any open HDF5 files to not trigger any `OSError`s, close running clusters
        # and clean up tmp dirs and created directories/log-files
        sigData.file.close()
        try:
            os.unlink(logFile)
        except PermissionError:
            pass
        shutil.rmtree(tempDir, ignore_errors=True)
        shutil.rmtree(tempDir2, ignore_errors=True)
        for folder in outDirs:
            shutil.rmtree(folder, ignore_errors=True)

        # Wait a second (literally) so that no new parallel workers started by
        # `test_existing_cluster` erroneously use existing HDF files
        time.sleep(1.0)

        return testclient

    # More functionality tests: ensure user-funcs w/multiple outputs are processed correctly
    def test_medium_filter(self, testclient=None):

        # Prepare data containers
        _, sigName = self._prep_data("acme_tmp2")

        # Collect auto-generated output directories in list for later cleanup
        outDirs = []

        # Parallelize across channels, write results to disk
        with ParallelMap(lowpass_medium,
                         sigName,
                         range(self.nChannels),
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
             pmap.compute()
        outDirs.append(pmap.out_dir)

        # Ensure output container was created correctly
        payloadDir = pmap.results_container.replace(".h5", "_payload")
        assert len(glob(os.path.join(payloadDir, "*.h5"))) == pmap.n_calls

        # Compare computed results to expected values
        with h5py.File(pmap.results_container, "r") as h5col:
            dset = "comp_{}/result_{}"
            for chNo in range(self.nChannels):
                assert len(h5col["comp_{}".format(chNo)].keys()) == 4
                assert np.mean(np.abs(h5col[dset.format(chNo, 0)][()] - self.orig[:, chNo])) < self.tol
                assert h5col[dset.format(chNo, 1)][()] == chNo
                assert np.array_equal(h5col[dset.format(chNo, 2)][()], self.b)
                assert np.array_equal(h5col[dset.format(chNo, 3)][()], self.a)

        # Remember results for later use
        colRes = str(pmap.results_container)

        # Same with `single_file`
        with ParallelMap(lowpass_medium,
                         sigName,
                         range(self.nChannels),
                         partition=defaultQ,
                         setup_interactive=False,
                         single_file=True) as pmap:
            pmap.compute()
        outDirs.append(pmap.out_dir)

        # Ensure only one file was generated
        assert glob(os.path.join(pmap.out_dir, "*")) == [pmap.results_container]

        # Compare single file to link collection computed above
        with h5py.File(colRes, "r") as h5col:
            with h5py.File(pmap.results_container, "r") as h5single:
                dset = "comp_{}/result_{}"
                for chNo in range(self.nChannels):
                    assert len(h5single["comp_{}".format(chNo)].keys()) == 4
                    assert np.array_equal(h5single[dset.format(chNo, 0)][()], h5col[dset.format(chNo, 0)][()])
                    assert h5col[dset.format(chNo, 1)][()] == h5single[dset.format(chNo, 1)][()]
                    assert np.array_equal(h5col[dset.format(chNo, 2)][()], h5single[dset.format(chNo, 2)][()])
                    assert np.array_equal(h5col[dset.format(chNo, 3)][()], h5single[dset.format(chNo, 3)][()])

        # Same, but collect results in memory: ensure nothing freaky happens
        with ParallelMap(lowpass_medium,
                         sigName,
                         range(self.nChannels),
                         write_worker_results=False,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            resInMem = pmap.compute()

        # Be double-paranoid: ensure on-disk and in-memory results match up
        with h5py.File(colRes, "r") as h5col:
            dset = "comp_{}/result_{}"
            for chNo in range(self.nChannels):
                assert len(resInMem[chNo]) == 4
                assert np.array_equal(h5col[dset.format(chNo, 0)][()], resInMem[chNo][0])
                assert h5col[dset.format(chNo, 1)][()] == resInMem[chNo][1]
                assert np.array_equal(h5col[dset.format(chNo, 2)][()], resInMem[chNo][2])
                assert np.array_equal(h5col[dset.format(chNo, 3)][()], resInMem[chNo][3])

        # Clean up created results directories
        for folder in outDirs:
            shutil.rmtree(folder, ignore_errors=True)

        # Wait a second (literally) so that no new parallel workers started by
        # `test_existing_cluster` erroneously use existing HDF files
        time.sleep(1.0)

        return testclient

    # Even more functionality tests: ensure output array stacking works
    def test_outshape(self, testclient=None):

        # Prepare data containers
        _, sigName = self._prep_data("acme_tmp_outshape")

        # Collect auto-generated output directories in list for later cleanup
        outDirs = []

        # Compute result length needed to determine final shape
        nSamples = self.fs * self.nTrials

        # Parallelize across channels, write results to disk
        with ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         result_shape=(None, nSamples),
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            resOnDisk = pmap.compute()
        outDirs.append(pmap.out_dir)

        # Remember results for later use
        colRes = str(pmap.results_container)

        # Compare computed single-channel results to expected low-freq signal
        # and ensure collection container was assembled correctly
        with h5py.File(pmap.results_container, "r") as h5col:
            assert len(h5col.keys()) == 1
            assert h5col["result_0"].is_virtual
            for chNo, h5name in enumerate(resOnDisk):
                with h5py.File(h5name, "r") as h5f:
                    assert np.mean(np.abs(h5f["result_0"][()] - self.orig[:, chNo])) < self.tol
                    assert np.array_equal(h5col["result_0"][chNo, :], h5f["result_0"][()])

        # Same but don't use a virtual dataset and transpose the final array
        with ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         result_shape=(nSamples, None),
                         single_file=True,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            pmap.compute()
        outDirs.append(pmap.out_dir)

        # Ensure only one file was generated
        assert glob(os.path.join(pmap.out_dir, "*")) == [pmap.results_container]

        # Compare single file to virtual dataset
        with h5py.File(colRes, "r") as h5col:
            with h5py.File(pmap.results_container, "r") as h5single:
                assert len(h5single.keys()) == 1
                assert h5single["result_0"].is_virtual is False
                assert np.array_equal(h5single["result_0"][()].T, h5col["result_0"][()])

        # Finally, ensure in-memory results-collection works as expected
        with ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         result_shape=(None, nSamples),
                         write_worker_results=False,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            resInMem = pmap.compute()
        with h5py.File(colRes, "r") as h5col:
            assert np.array_equal(h5col["result_0"][()], resInMem[0])

        # Ensure dtype is respected
        with ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         result_shape=(None, nSamples),
                         result_dtype="float16",
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            pmap.compute()
        outDirs.append(pmap.out_dir)
        with h5py.File(pmap.results_container, "r") as h5f:
            assert h5f["result_0"].dtype.name == "float16"

        # Ensure invalid dtypes don't pass through
        with pytest.raises(TypeError) as tperr:
            with ParallelMap(lowpass_simple,
                             sigName,
                             range(self.nChannels),
                             result_shape=(None, nSamples),
                             result_dtype=np.ones((3,)),
                             partition=defaultQ,
                             setup_interactive=False) as pmap:
                pmap.compute()
            assert "`result_dtype` has to be a string" in str(tperr)
        with pytest.raises(TypeError) as tperr:
            with ParallelMap(lowpass_simple,
                             sigName,
                             range(self.nChannels),
                             result_shape=(None, nSamples),
                             result_dtype="invalid",
                             partition=defaultQ,
                             setup_interactive=False) as pmap:
                pmap.compute()
            assert "`result_dtype` has to be a valid NumPy datatype" in str(tperr)

        # Ensure borked shapes are caught
        with pytest.raises(TypeError) as tperr:
            with ParallelMap(lowpass_simple,
                             sigName,
                             range(self.nChannels),
                             result_shape=3,
                             partition=defaultQ,
                             setup_interactive=False) as pmap:
                pmap.compute()
            assert "`result_shape` has to be either `None` or tuple" in str(tperr)
        with pytest.raises(ValueError) as valerr:
            with ParallelMap(lowpass_simple,
                             sigName,
                             range(self.nChannels),
                             result_shape=(None, None, nSamples),
                             partition=defaultQ,
                             setup_interactive=False) as pmap:
                pmap.compute()
            assert "`result_shape` must contain exactly one `None`" in str(valerr)
        with pytest.raises(ValueError) as valerr:
            with ParallelMap(lowpass_simple,
                             sigName,
                             range(self.nChannels),
                             result_shape=(3, nSamples),
                             partition=defaultQ,
                             setup_interactive=False) as pmap:
                pmap.compute()
            assert "`result_shape` must contain exactly one `None`" in str(valerr)
        with pytest.raises(ValueError) as valerr:
            with ParallelMap(lowpass_simple,
                             sigName,
                             range(self.nChannels),
                             result_shape=("invalid", None, nSamples),
                             partition=defaultQ,
                             setup_interactive=False) as pmap:
                pmap.compute()
            assert "`result_shape` must only contain numerical values" in str(valerr)
        with pytest.raises(ValueError) as valerr:
            with ParallelMap(lowpass_simple,
                             sigName,
                             range(self.nChannels),
                             result_shape=(-3, None, nSamples),
                             partition=defaultQ,
                             setup_interactive=False) as pmap:
                pmap.compute()
            assert "`result_shape` must only contain non-negative integers" in str(valerr)
        with pytest.raises(ValueError) as valerr:
            with ParallelMap(lowpass_simple,
                             sigName,
                             range(self.nChannels),
                             result_shape=(np.pi, None, nSamples),
                             partition=defaultQ,
                             setup_interactive=False) as pmap:
                pmap.compute()
            assert "`result_shape` must only contain non-negative integers" in str(valerr)

        # Emergency pickling
        with ParallelMap(pickle_func,
                         self.sig,
                         self.b,
                         self.a,
                         range(self.nChannels),
                         sabotage_hdf5=True,
                         n_inputs=self.nChannels,
                         result_shape=(nSamples, None),
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            mixedResults = pmap.compute()
        outDirs.append(pmap.out_dir)

        # Ensure pickles and hdf5's live together happily
        resultsContainer = os.path.basename(mixedResults[0])
        resultsContainer = os.path.join(os.path.dirname(mixedResults[0]),
                                        resultsContainer[:resultsContainer.rfind("_0")] + ".h5")
        payloadDir = resultsContainer.replace(".h5", "_payload")
        assert pmap.results_container is None
        assert not os.path.isfile(resultsContainer)
        assert not os.path.isdir(payloadDir)
        assert all(os.path.isfile(fle) for fle in mixedResults)
        assert len(mixedResults) == pmap.n_calls

        # Ensure deliberate pickling doesn't clash w/(erroneous) shape spec
        with ParallelMap(pickle_func,
                         self.sig,
                         self.b,
                         self.a,
                         range(self.nChannels),
                         sabotage_hdf5=False,
                         n_inputs=self.nChannels,
                         result_shape=(nSamples, None),
                         write_pickle=True,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            pickles = pmap.compute()
        outDirs.append(pmap.out_dir)
        assert all(os.path.isfile(fle) for fle in pickles)
        assert len(pickles) == pmap.n_calls

        # Ensure multiple return values are handled correctly
        with ParallelMap(lowpass_medium,
                         sigName,
                         range(self.nChannels),
                         result_shape=(None, nSamples),
                         single_file=False,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
             pmap.compute()
        outDirs.append(pmap.out_dir)

        # Save results for later
        multiRet = str(pmap.results_container)

        # Compare computed results to stored "reference"
        with h5py.File(colRes, "r") as h5col:
            with h5py.File(pmap.results_container, "r") as h5f:
                assert len(h5f.keys()) == pmap.n_calls + 1
                assert h5f["result_0"].is_virtual is True
                assert np.array_equal(h5f["result_0"][()], h5col["result_0"][()])
                for k in range(pmap.n_calls):
                    assert len(h5f["comp_{}".format(k)].keys()) == 3
                    assert h5f["comp_{}/{}".format(k, "result_1")][()] == k
                    assert np.array_equal(h5f["comp_{}/{}".format(k, "result_2")][()], self.b)
                    assert np.array_equal(h5f["comp_{}/{}".format(k, "result_3")][()], self.a)

        # Same w/single output container
        with ParallelMap(lowpass_medium,
                         sigName,
                         range(self.nChannels),
                         result_shape=(None, nSamples),
                         single_file=True,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
             pmap.compute()
        outDirs.append(pmap.out_dir)

        # Compare results
        with h5py.File(multiRet, "r") as h5ref:
            with h5py.File(pmap.results_container, "r") as h5f:
                assert len(h5f.keys()) == pmap.n_calls + 1
                for k in range(pmap.n_calls):
                    for rk in range(1,4):
                        dset = "comp_{}/result_{}".format(k, rk)
                        assert np.array_equal(h5f[dset], h5ref[dset])

        # Finally, ensure in-memory results-collection works w/multiple returns
        with ParallelMap(lowpass_medium,
                         sigName,
                         range(self.nChannels),
                         result_shape=(None, nSamples),
                         write_worker_results=False,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            resInMem = pmap.compute()
        with h5py.File(multiRet, "r") as h5ref:
            assert np.array_equal(h5ref["result_0"][()], resInMem[0])
            rCount = 1
            for k in range(pmap.n_calls):
                assert h5ref["comp_{}/{}".format(k, "result_1")][()] == resInMem[rCount]
                rCount +=1
                assert np.array_equal(h5ref["comp_{}/{}".format(k, "result_2")][()], resInMem[rCount])
                rCount +=1
                assert np.array_equal(h5ref["comp_{}/{}".format(k, "result_3")][()], resInMem[rCount])
                rCount +=1

        # Clean up created results directories
        for folder in outDirs:
            shutil.rmtree(folder, ignore_errors=True)

        # Wait a second (literally) so that no new parallel workers started by
        # `test_existing_cluster` erroneously use existing HDF files
        time.sleep(1.0)

        return testclient

    # Test if pickling/emergency pickling and I/O in general works as intended
    def test_pickling(self, testclient=None):

        # Collected auto-generated output directories in list for later cleanup
        outDirs = []

        # Execute `pickle_func` w/regular HDF5 saving
        with ParallelMap(pickle_func,
                         self.sig,
                         self.b,
                         self.a,
                         range(self.nChannels),
                         sabotage_hdf5=False,
                         n_inputs=self.nChannels,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            hdfResults = pmap.compute()
        colRes = str(pmap.results_container)
        outDirs.append(pmap.out_dir)

        # Execute `pickle_func` w/pickling
        with ParallelMap(pickle_func,
                         self.sig,
                         self.b,
                         self.a,
                         range(self.nChannels),
                         n_inputs=self.nChannels,
                         write_pickle=True,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            pklResults = pmap.compute()
        outDirs.append(pmap.out_dir)

        # Ensure HDF5 and pickle match up
        with h5py.File(colRes, "r") as h5col:
            dset = "comp_{}/result_0"
            for chNo in range(self.nChannels):
                with open(pklResults[chNo], "rb") as pkf:
                    pklRes = pickle.load(pkf)
                assert np.array_equal(pklRes, h5col[dset.format(chNo)][()])

        # Ensure single_file and pickling does not work
        with pytest.raises(ValueError) as valerr:
            with ParallelMap(pickle_func,
                            self.sig,
                            self.b,
                            self.a,
                            range(self.nChannels),
                            n_inputs=self.nChannels,
                            write_pickle=True,
                            single_file=True,
                            partition=defaultQ,
                            setup_interactive=False) as pmap:
                pmap.compute()
            assert "Pickling of results does not support single output file creation" in str(valerr.value)

        # Test emergency pickling
        with ParallelMap(pickle_func,
                         self.sig,
                         self.b,
                         self.a,
                         range(self.nChannels),
                         sabotage_hdf5=True,
                         n_inputs=self.nChannels,
                         partition=defaultQ,
                         setup_interactive=False) as pmap:
            mixedResults = pmap.compute()
        outDirs.append(pmap.out_dir)

        # Collection container should have been auto-removed
        resultsContainer = os.path.basename(mixedResults[0])
        resultsContainer = os.path.join(os.path.dirname(mixedResults[0]),
                                        resultsContainer[:resultsContainer.rfind("_0")] + ".h5")
        payloadDir = resultsContainer.replace(".h5", "_payload")
        assert pmap.results_container is None
        assert not os.path.isfile(resultsContainer)
        assert not os.path.isdir(payloadDir)

        # Ensure non-compliant dicts were pickled, rest is in HDF5
        for chNo, fname in enumerate(mixedResults):
            if chNo % 2 == 0:
                assert fname.endswith(".pickle")
                with open(fname, "rb") as pkf:
                    assert np.array_equal(self.b, pickle.load(pkf)[0]["b"])
            else:
                assert fname.endswith(".h5")
                with h5py.File(fname, "r") as h5f:
                    with h5py.File(hdfResults[chNo], "r") as h5ref:
                        assert np.array_equal(h5f["result_0"][()], h5ref["result_0"][()])

        # Ensure emergency pickling and single file does not work
        with pytest.raises(RuntimeError):
            with ParallelMap(pickle_func,
                            self.sig,
                            self.b,
                            self.a,
                            range(self.nChannels),
                            sabotage_hdf5=True,
                            n_inputs=self.nChannels,
                            single_file=True,
                            partition=defaultQ,
                            setup_interactive=False) as pmap:
                pmap.compute()


        # Test write breakdown (both for HDF5 saving and pickling)
        pmap = ParallelMap(pickle_func,
                           self.sig,
                           self.b,
                           self.a,
                           range(self.nChannels),
                           sabotage_hdf5=True,
                           n_inputs=self.nChannels,
                           partition=defaultQ,
                           setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)
        pmap.kwargv["outFile"][0] = "/path/to/nowhere"
        with pytest.raises(RuntimeError) as runerr:
            pmap.compute()
            assert "<ACMEdaemon> Parallel computation failed" in str(runerr.value)
        pmap = ParallelMap(pickle_func,
                           self.sig,
                           self.b,
                           self.a,
                           range(self.nChannels),
                           sabotage_hdf5=True,
                           n_inputs=self.nChannels,
                           write_pickle=True,
                           partition=defaultQ,
                           setup_interactive=False)
        outDirs.append(pmap.daemon.out_dir)
        pmap.kwargv["outFile"][0] = "/path/to/nowhere"
        with pytest.raises(RuntimeError) as runerr:
            pmap.compute()
            assert "<ACMEdaemon> Parallel computation failed" in str(runerr.value)

        # Clean up testing folder
        for folder in outDirs:
            shutil.rmtree(folder, ignore_errors=True)

        return testclient

    # test if KeyboardInterrupts are handled correctly
    @skip_if_not_linux
    def test_cancel(self):

        # Setup temp-directory layout for subprocess-scripts and prepare interpreters
        tempDir = os.path.join(os.path.abspath(os.path.expanduser("~")), "acme_tmp")
        os.makedirs(tempDir, exist_ok=True)
        pshells = [os.path.join(os.path.split(sys.executable)[0], pyExec) for pyExec in ["python", "ipython"]]

        # Prepare ad-hoc script for execution in new process
        scriptName = os.path.join(tempDir, "dummy.py")
        scriptContents = \
            "from acme import ParallelMap, cluster_cleanup\n" +\
            "import time\n" +\
            "def long_running(dummy):\n" +\
            "   time.sleep(10)\n" +\
            "   return\n" +\
            "if __name__ == '__main__':\n" +\
            "   cluster_cleanup() \n" +\
            "   with ParallelMap(long_running, [None]*2, setup_interactive=False, partition='8GBXS', write_worker_results=False) as pmap: \n" +\
            "       pmap.compute()\n" +\
            "   print('ALL DONE')\n"
        with open(scriptName, "w") as f:
            f.write(scriptContents)

        # Execute the above script both in Python and iPython to ensure global functionality
        for pshell in pshells:

            # Launch new process in background (`stdbuf` prevents buffering of stdout)
            proc = subprocess.Popen("stdbuf -o0 " + pshell + " " + scriptName,
                                    shell=True, start_new_session=True,
                                    stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=0)

            # Wait for ACME to start up (as soon as logging info is shown, `pmap.compute()` is running)
            # However: don't wait indefinitely - if `pmap.compute` is not started within 30s, abort
            logStr = "<ParallelMap> INFO: Log information available at"
            buffer = bytearray()
            timeout = 30
            t0 = time.time()
            for line in itertools.takewhile(lambda x: time.time() - t0 < timeout, iter(proc.stdout.readline, b"")):
                buffer.extend(line)
                if logStr in line.decode("utf8"):
                    break
            assert logStr in buffer.decode("utf8")

            # Wait a bit, then simulate CTRL+C in sub-process; make sure the above
            # impromptu script did not run to completion *but* the created client was
            # shut down with CTRL + C
            time.sleep(2)
            os.killpg(proc.pid, sys_signal.SIGINT)
            time.sleep(1)
            out = proc.stdout.read().decode()
            assert "ALL DONE" not in out
            assert "INFO: <cluster_cleanup> Successfully shut down" in out

        # Almost identical script, this time use an externally started client
        scriptName = os.path.join(tempDir, "dummy2.py")
        scriptContents = \
            "from acme import ParallelMap, esi_cluster_setup\n" +\
            "import time\n" +\
            "def long_running(dummy):\n" +\
            "   time.sleep(10)\n" +\
            "   return\n" +\
            "if __name__ == '__main__':\n" +\
            "   client = esi_cluster_setup(partition='8GBDEV',n_workers=1, interactive=False)\n" +\
            "   with ParallelMap(long_running, [None]*2, setup_interactive=False, write_worker_results=False) as pmap: \n" +\
            "       pmap.compute()\n" +\
            "   print('ALL DONE')\n"
        with open(scriptName, "w") as f:
            f.write(scriptContents)

        # Test script functionality in both Python and iPython
        for pshell in pshells:
            proc = subprocess.Popen("stdbuf -o0 " + sys.executable + " " + scriptName,
                                    shell=True, start_new_session=True,
                                    stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=0)
            logStr = "<ParallelMap> INFO: Log information available at"
            buffer = bytearray()
            timeout = 60
            t0 = time.time()
            for line in itertools.takewhile(lambda x: time.time() - t0 < timeout, iter(proc.stdout.readline, b"")):
                buffer.extend(line)
                if logStr in line.decode("utf8"):
                    break
            assert logStr in buffer.decode("utf8")
            time.sleep(2)
            os.killpg(proc.pid, sys_signal.SIGINT)
            time.sleep(2)
            out = proc.stdout.read().decode()
            assert "ALL DONE" not in out
            assert "<ParallelMap> INFO: <ACME> CTRL + C acknowledged, client and workers successfully killed" in out

        # Ensure random exception does not immediately kill an active client
        scriptName = os.path.join(tempDir, "dummy3.py")
        scriptContents = \
            "from acme import esi_cluster_setup\n" +\
            "import time\n" +\
            "if __name__ == '__main__':\n" +\
            "   esi_cluster_setup(partition='8GBDEV',n_workers=1, interactive=False)\n" +\
            "   time.sleep(60)\n"
        with open(scriptName, "w") as f:
            f.write(scriptContents)
        proc = subprocess.Popen("stdbuf -o0 " + sys.executable + " " + scriptName,
                                shell=True, start_new_session=True,
                                stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=0)

        # Give the client time to start up, then send a floating-point exception
        # (equivalent to a `ZeroDivsionError` to the child process)
        time.sleep(5)
        assert proc.poll() is None
        proc.send_signal(sys_signal.SIGFPE)

        # Ensure the `ZeroDivsionError` did not kill the process. Then terminate it
        # and confirm that the floating-exception was propagated correctly
        assert proc.poll() is None
        proc.terminate()
        proc.wait()
        assert proc.returncode in [-sys_signal.SIGFPE.value, -sys_signal.SIGTERM.value]

        # Clean up tmp folder
        shutil.rmtree(tempDir, ignore_errors=True)

    # test dryrun keyword
    def test_dryrun(self, monkeypatch):

        # This call tests two things: first, the actual dryrun must work, second
        # using pytest's monkeypatch fixture user-input is simulated. If a user
        # decides to not move ahead after the dryrun the auto-generated output
        # directory must be cleaned up
        monkeypatch.setattr("builtins.input", lambda _ : "n")
        pmap = ParallelMap(simple_func, [2, 4, 6, 8], 4, setup_interactive=True, dryrun=True)

        # Ensure auto-generated output dir has been successfully removed
        outDir = pmap.daemon.out_dir
        assert os.path.exists(outDir) is False

    def test_memest(self):

        # First kill any existing client (otherwise profiling is not happening)
        cluster_cleanup()

        # Create tmp directory for logfile
        tempDir = os.path.join(os.path.abspath(os.path.expanduser("~")), "acme_tmp")
        os.makedirs(tempDir, exist_ok=True)
        customLog = os.path.join(tempDir, "acme_log.txt")
        outDirs = []

        # Set `arrsize` depending on available runner hardware and prepare expected
        # mem estimates accordingly
        arrsize = 2
        estMem = 3
        if platform.machine() == "ppc64le":
            arrsize = 0.5
            estMem = 1

        # Prepare `ParallelMap` instance for 2 concurrent calls of `memtest_func`:
        # a 2GB array is allocated, set final sleep wait period to 2 seconds, so
        # total runtime of the function should be b/w 5-10 seconds (depending on how long
        # array allocation takes)
        pmap = ParallelMap(memtest_func,
                           np.arange(2),
                           2,
                           sleeper=2,
                           arrsize=arrsize,
                           logfile=customLog,
                           setup_interactive=False)

        # If executed locally, the above call did not invoke `estimate_memuse` since
        # there's no partition to choose
        if not useSLURM:

            # Fire off `estimate_memuse` manually
            tic = time.perf_counter()
            memEst = pmap.daemon.estimate_memuse()
            toc = time.perf_counter()

            # Ensure 2.12GB array allocation was profiled correctly (2.x GB are rounded
            # upwards towards the next integer, i.e., 2.13 GB -> estimate_memuse:3)
            if sys.platform != "darwin":
                assert memEst == "estimate_memuse:" + str(estMem)

            # Ensure profiler quit out early (since total runtime of `memtest_func` is below 30s)
            assert toc - tic < 30

        # Check that for 2 parallel calls only 2 workers were memory-profiled
        with open(customLog, "r", encoding="utf8") as f:
            logTxt = f.read()
        assert "Estimated memory consumption across 2 runs" in logTxt

        # If running on the ESI cluster, ensure the correct partition has been picked
        if onESI and useSLURM:
            assert "Picked partition 8GBXS based on estimated memory consumption of 3 GB" in logTxt

        # Profiling completed full run of `memtest_func`: ensure any auto-created
        # output HDF5 files were removed
        outDirs.append(pmap.daemon.out_dir)
        assert len(os.listdir(os.path.dirname(pmap.kwargv['outFile'][0]))) == 0

        # Syncopy-related test: ensure memory profiling works also if ACME
        # does not handle result collection
        syncopylog = os.path.join(tempDir, "syncopylog.txt")
        pmap = ParallelMap(memtest_func,
                           np.arange(100),
                           2,
                           sleeper=300,
                           arrsize=arrsize,
                           logfile=syncopylog,
                           setup_timeout=10,
                           write_worker_results=False,
                           setup_interactive=False)
        with open(syncopylog, "r", encoding="utf8") as f:
            logTxt = f.read()
        assert "memEstRun" not in logTxt

        # Now prepare `ParallelMap` instance for 100 concurrent calls of `memtest_func`:
        # again a 2GB array is allocated, but set final sleep to 5 minutes, enforcing
        # abort of profiling runs (set `setup_interactive` low to not have 100
        # SLURM workers being actually started)
        del pmap
        cluster_cleanup()
        customLog2 = os.path.join(tempDir, "acme_log2.txt")
        pmap = ParallelMap(memtest_func,
                           np.arange(100),
                           2,
                           sleeper=300,
                           arrsize=arrsize,
                           logfile=customLog2,
                           setup_timeout=10,
                           setup_interactive=False)

        # Again, fire off `estimate_memuse` manually if tests are run locally
        if not useSLURM:
            tic = time.perf_counter()
            memEst = pmap.daemon.estimate_memuse()
            toc = time.perf_counter()

            # Ensure 2.12GB array allocation was profiled correctly
            if sys.platform != "darwin":
                assert memEst == "estimate_memuse:" + str(estMem)

            # Ensure profiler went the whole mile (since total runtime of `memtest_func` > 5 min)
            assert 140 < toc - tic < 200

        # Check that for max 5 workers were memory-profiled
        with open(customLog2, "r", encoding="utf8") as f:
            logTxt = f.read()
        assert "Estimated memory consumption across 5 runs" in logTxt

        # If running on the ESI cluster, ensure the correct partition has been picked (again)
        if onESI and useSLURM:
            assert "Picked partition 8GBXS based on estimated memory consumption of 3 GB" in logTxt

        # Profiling should not have generated any output
        outDirs.append(pmap.daemon.out_dir)
        assert len(os.listdir(os.path.dirname(pmap.kwargv['outFile'][0]))) == 0

        # Prepare final "full" tests
        del pmap
        cluster_cleanup()
        customLog3 = os.path.join(tempDir, "acme_log3.txt")

        # Assert that `partition="auto"` has no effect in `LocalCluster` case
        if not useSLURM:
            with ParallelMap(memtest_func,
                             np.arange(2),
                             2,
                             sleeper=2,
                             arrsize=arrsize,
                             logfile=customLog3) as pmap:
                pmap.compute()
            with open(customLog3, "r", encoding="utf8") as f:
                logTxt = f.read()
            assert "Estimating memory consumption" not in logTxt
            outDirs.append(pmap.out_dir)

        else:

            # Simulate call of ParallelMap(partition="auto",...) but w/wrong mem_per_worker!
            with pytest.raises(customIOError):
                esi_cluster_setup(partition="auto", mem_per_worker="invalid")

            # Simulate `ParallelMap(partition="auto",...)` call by invoking `esi_cluster_setup`
            # with `mem_per_worker='esstimate_memuse:XY'`
            client = esi_cluster_setup(partition="auto",
                                       mem_per_worker="estimate_memuse:12",
                                       n_workers=1,
                                       interactive=False)

            # Ensure the right partition was picked (16GBXY, not 8GBXY)
            assert "16GB" in client.cluster.job_header.split("-p ")[1].split("\n")[0]
            cluster_cleanup(client)

            # Full run (finally) w/10 workers, 5 of em get mem-profiled
            with ParallelMap(memtest_func,
                             np.arange(10),
                             2,
                             sleeper=35,
                             arrsize=arrsize,
                             partition="auto",
                             logfile=customLog3,
                             setup_interactive=False) as pmap:
                pmap.compute()

            # Check correct partition and no. of workers profiled
            with open(customLog3, "r", encoding="utf8") as f:
                logTxt = f.read()
            assert "Estimated memory consumption across 5 runs" in logTxt
            assert "Picked partition 8GBXS" in logTxt
            outDirs.append(pmap.out_dir)

        # Clean up
        shutil.rmtree(tempDir, ignore_errors=True)
        for folder in outDirs:
            shutil.rmtree(folder, ignore_errors=True)
        time.sleep(0.1)

    # test if deprecated keywords are mapped onto new names correctly
    def test_backcompat(self):

        # Prepare data containers
        _, sigName = self._prep_data("acme_tmp")

        # Collect auto-generated output directories in list for later cleanup
        outDirs = []

        # Invoke `ParallelMap` w/deprecated `n_jobs` kw
        with ParallelMap(lowpass_simple,
                         sigName,
                         range(self.nChannels),
                         partition=defaultQ,
                         n_jobs=2,
                         logfile=True,
                         setup_interactive=False) as pmap:
            pmap.compute()
        outDirs.append(pmap.out_dir)

        # Ensure a deprecation warning was issued
        for handler in pmap.log.handlers:
            if isinstance(handler, logging.FileHandler):
                with open(handler.baseFilename, "r") as fl:
                    logTxt = fl.read()
                assert "DEPRECATED" in logTxt

        if useSLURM:

            # Use the deprecated `n_jobs` and `mem_per_job` keywords
            n_workers = 2
            mem_per_worker = "2GB"
            with ParallelMap(lowpass_simple,
                            sigName,
                            range(self.nChannels),
                            partition=defaultQ,
                            n_jobs=n_workers,
                            mem_per_job=mem_per_worker,
                            stop_client=False,
                            setup_interactive=False) as pmap:
                pmap.compute()
            outDirs.append(pmap.out_dir)

            # Ensure the provided input was interpreted correctly
            client = dd.get_client()
            assert pmap.n_workers == n_workers
            assert len(client.cluster.workers) == pmap.n_workers
            memory = np.unique([w["memory_limit"] for w in client.cluster.scheduler_info["workers"].values()])
            assert memory.size == 1
            assert round(memory[0] / 1000**3) == int(mem_per_worker.replace("GB", ""))

        # Clean up
        for folder in outDirs:
            shutil.rmtree(folder, ignore_errors=True)
        time.sleep(0.1)
        cluster_cleanup()

    # test esi-cluster-setup called separately before pmap
    def test_existing_cluster(self):

        # Test custom SLURM cluster setup
        if useSLURM:

            # Supply extra args to start client for actual tests
            slurmOut = "/cs/home/{}/acme_out".format(getpass.getuser())
            client = esi_cluster_setup(partition=defaultQ,
                                       n_workers=10,
                                       job_extra=["--output={}".format(slurmOut)],
                                       interactive=False)
            assert "--output={}".format(slurmOut) in client.cluster.job_header

        else:
            client = esi_cluster_setup(interactive=False)

        # Re-run tests with pre-allocated client (except for those in `skipTests`); ensure
        # client "survives" multiple independent test runs and is not accidentally closed
        skipTests = ["test_existing_cluster", "test_cancel", "test_dryrun",
                     "test_memest", "test_backcompat", "_prep_data"]
        all_tests = [attr for attr in self.__dir__()
                     if (inspect.ismethod(getattr(self, attr)) and attr not in skipTests)]
        for test in all_tests:
            clnt = getattr(self, test)(testclient=client)
            assert clnt == client
        client.close()
        client.cluster.close()
        if useSLURM:
            shutil.rmtree(slurmOut, ignore_errors=True)



# -*- coding: utf-8 -*-
#
# Simple script for testing acme w/o pip-installing it
#

# Builtin/3rd party package imports
import numpy as np

# Add acme to Python search path
import os
import sys
acme_path = os.path.abspath(".." + os.sep + "..")
if acme_path not in sys.path:
    sys.path.insert(0, acme_path)

# Import package
from acme import ParallelMap

def f(x, y, z=3, w=np.zeros((3, 1)), **kwargs):
    return (x + y) * z * w.max()

import time

def g(x, y, z=3):
    fSize = np.dtype("float").itemsize
    arrSize = 4
    time.sleep(10)
    arr = np.ones((int(arrSize * 1024**3 / fSize), ))
    time.sleep(300)
    return (sum(x) + y) * z * arr.max()

def arr_test(x, y):
    return x + y

# Prepare code to be executed using, e.g., iPython's `%run` magic command
if __name__ == "__main__":

    # Test stuff within here...
    # pass

    with ParallelMap(arr_test, [np.ones((20,)), 2 * np.ones((20,)), 3 * np.ones((20,))], 4, result_shape=(None, 20)) as pmap:
        results = pmap.compute()

    # pmap = ParallelMap(g, np.arange(100), 2)
    # pmap.daemon.estimate_memuse()

    # # with pmap as p:
    # #     p.compute()





# Builtins
import datetime
from setuptools import setup
from setuptools.config.setupcfg import read_configuration
import subprocess
import toml

# External packages
import yaml
from setuptools_scm import get_version

# Set release version by hand
releaseVersion = "2022.12"

# Read dependencies from setup.cfg + pyproject.toml and create conda environment file
envFile = "acme.yml"
setupOpts = read_configuration("setup.cfg")["options"]
tomlPkgs = toml.load("pyproject.toml")["build-system"]["requires"]
allPkgs = setupOpts["install_requires"] + setupOpts["extras_require"]["dev"] + tomlPkgs
pipPkgs = ["dask-jobqueue", "sphinx_automodapi"]
for k in range(len(pipPkgs)):
    pkg = pipPkgs[k]
    pipPkgs[k] = allPkgs.pop([allPkgs.index(dep) for dep in allPkgs if pkg in dep][0])
pyVer = setupOpts
allPkgs += ["python " + str(setupOpts["python_requires"]), "pip", {"pip" : pipPkgs}]
ymlDict = {"name" : "acme",
           "channels" : ["defaults", "conda-forge"],
           "dependencies" : allPkgs}
msg = "# This file was auto-generated by setup.py on {}. \n" +\
    "# Do not edit, all of your changes will be overwritten. \n"
msg = msg.format(datetime.datetime.now().strftime("%d/%m/%Y at %H:%M:%S"))
with open(envFile, 'w') as ymlFile:
    ymlFile.write(msg)
    yaml.dump(ymlDict, ymlFile, default_flow_style=False)

# If code has not been obtained via `git` or we're inside the main branch,
# use the hard-coded `releaseVersion` as version. Otherwise keep the local `tag.devx`
# scheme for TestPyPI uploads
proc = subprocess.run("git branch --show-current", shell=True, capture_output=True, text=True)
if proc.returncode !=0 or proc.stdout.strip() == "main":
    version = releaseVersion
    versionKws = {"use_scm_version" : False, "version" : version}
else:
    version = get_version(root='.', relative_to=__file__, local_scheme="no-local-version")
    versionKws = {"use_scm_version" : {"local_scheme": "no-local-version"}}

# Update citation file
citationFile = "CITATION.cff"
with open(citationFile) as ymlFile:
    ymlObj = yaml.safe_load(ymlFile)
ymlObj["version"] = version
ymlObj["date-released"] = datetime.datetime.now().strftime("%Y-%m-%d")
with open(citationFile, "w") as ymlFile:
    yaml.dump(ymlObj, ymlFile)

# Run setup (note: identical arguments supplied in setup.cfg will take precedence)
setup(
    setup_requires=['setuptools_scm'],
    **versionKws
)



# -*- coding: utf-8 -*-
#
# Computational scaffolding for user-interface
#

# Builtin/3rd party package imports
import time
import socket
import getpass
import datetime
import inspect
import numbers
import collections
import os
import sys
import glob
import shutil
import pickle
import logging
import functools
import multiprocessing
import psutil
import tqdm
import h5py
import dask
import dask.distributed as dd
import dask_jobqueue as dj
import numpy as np

# Local imports
from . import __path__
from .dask_helpers import (esi_cluster_setup, local_cluster_setup,
                           slurm_cluster_setup, cluster_cleanup)
from .shared import user_yesno, is_esi_node
from . import shared as acs
isSpyModule = False
if "syncopy" in sys.modules:
    isSpyModule = True

__all__ = ["ACMEdaemon"]


# Main manager for parallel execution of user-defined functions
class ACMEdaemon(object):

    # Restrict valid class attributes
    __slots__ = "func", "argv", "kwargv", "n_calls", "n_workers", "acme_func", \
        "task_ids", "out_dir", "collect_results", "results_container", "result_shape", \
        "result_dtype", "stacking_dim", "client", "stop_client", "has_slurm", "log"

    # Prepend every stdout/stderr message with the name of this class
    msgName = "<ACMEdaemon>"

    # format string for tqdm progress bars
    tqdmFormat = "{desc}: {percentage:3.0f}% |{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}]"

    # time (in seconds) b/w querying state of futures ('pending' -> 'finished')
    sleepTime = 0.1

    def __init__(
        self,
        pmap=None,
        func=None,
        argv=None,
        kwargv=None,
        n_calls=None,
        n_workers="auto",
        write_worker_results=True,
        output_dir=None,
        result_shape=None,
        result_dtype="float",
        single_file=False,
        write_pickle=False,
        dryrun=False,
        partition="auto",
        mem_per_worker="auto",
        setup_timeout=60,
        setup_interactive=True,
        stop_client="auto",
        verbose=None,
        logfile=None,
        **kwargs):
        """
        Manager class for performing concurrent user function calls

        Parameters
        ----------
        pmap : :class:~`acme.ParallelMap` context manager or None
            If `pmap` is not `None`, `:class:~`acme.ACMEDaemon` assumes
            that that the provided :class:~`acme.ParallelMap` instance has already
            been properly set up to process `func` (all input arguments parsed and
            properly formatted). All other input arguments of `:class:~`acme.ACMEDaemon`
            are extracted from the provided :class:~`acme.ParallelMap` instance.
            If `pmap` is `None`, `:class:~`acme.ACMEDaemon` runs in "stand-alone"
            mode: all remaining arguments have to be manually supplied (in the
            correct format)
        func : callable
            User-defined function to be executed concurrently. See :class:~`acme.ParallelMap`
            for details.
        argv : list of lists
            Positional arguments of `func`: all elements of have to be list-like
            with lengths `n_calls` or 1
        kwargv : list of dicts
            Keyword arguments of `func`: all values of have to be list-like with
            lengths `n_calls` or 1
        n_calls : int
            Number of concurrent calls of `func` to perform. If `pmap` is not `None`,
            then ``n_calls = pmap.n_inputs``
        n_workers : int or "auto"
            Number of SLURM workers (=jobs) to spawn. See :class:~`acme.ParallelMap`
            for details.
        write_worker_results : bool
            If `True`, the return value(s) of `func` is/are saved on disk. See
            :class:~`acme.ParallelMap` for details.
        output_dir : str or None
            If provided, auto-generated results are stored in the given path. See
            :class:~`acme.ParallelMap` for details.
        result_shape : tuple or None
            If provided, results are slotted into a dataset/array with layout `result_shape`. See
            :class:~`acme.ParallelMap` for details.
        result_dtype : str or None
            If provided, determines numerical datatype of dataset laid out
            by `result_shape`. See :class:~`acme.ParallelMap` for details.
        single_file : bool
            If `True`, parallel workers write to the same results container. See
            :class:~`acme.ParallelMap` for details.
        write_pickle : bool
            If `True`, the return value(s) of `func` is/are pickled to disk. See
            :class:~`acme.ParallelMap` for details.
        dryrun : bool
            If `True`, a dry-run of calling `func` is performed using a single
            `args`, `kwargs` tuple. See :class:~`acme.ParallelMap` for details.
        partition : str
            Name of SLURM partition to use. See :class:~`acme.ParallelMap` for details.
        mem_per_worker : str
            Memory booking for each SLURM worker. See :class:~`acme.ParallelMap` for details.
        setup_timeout : int
            Timeout period (in seconds) for SLURM workers to come online. See
            :class:~`acme.ParallelMap` for details.
        setup_interactive : bool
            If `True`, user input is queried in case not enough SLURM workers could
            be started within `setup_timeout` seconds. See :class:~`acme.ParallelMap`
            for details.
        stop_client : bool or "auto"
            If `"auto"`, automatically started distributed computing clients
            are shut down at the end of computation, while user-provided clients
            are left untouched. See :class:~`acme.ParallelMap` for details.
        verbose : None or bool
            If `None` (default), general run-time information as well as warnings
            and errors are shown. See :class:~`acme.ParallelMap` for details.
        logfile : None or bool or str
            If `None` (default) or `False`, all run-time information as well as errors and
            warnings are printed to the command line only. See :class:~`acme.ParallelMap`
            for details.

        Returns
        -------
        results : list
            If `write_worker_results` is `True`, `results` is a list of HDF5 file-names
            containing computed results. If `write_worker_results` is `False`,
            results is a list comprising the actual return values of `func`.
            If `:class:~`acme.ACMEDaemon` was instantiated by :class:~`acme.ParallelMap`,
            results are propagated back to :class:~`acme.ParallelMap`.

        See also
        --------
        ParallelMap : Context manager and main user interface
        """

        # The only error checking happening in `__init__`
        if pmap is not None:
            if pmap.__class__.__name__ != "ParallelMap":
                msg = "{} `pmap` has to be a `ParallelMap` instance, not {}"
                raise TypeError(msg.format(self.msgName, str(pmap)))

        # Input pre-processed by a `ParallelMap` object takes precedence over keyword args
        self.initialize(getattr(pmap, "func", func),
                        getattr(pmap, "argv", argv),
                        getattr(pmap, "kwargv", kwargv),
                        getattr(pmap, "n_inputs", n_calls))

        # If `log` is `None`, `prepare_log` has not been called yet
        if getattr(pmap, "log", None) is None:
            self.log = acs.prepare_log(func, caller=self.msgName, logfile=logfile,
                                       verbose=verbose)
        else:
            self.log = pmap.log

        # Set up output handler
        self.prepare_output(write_worker_results,
                            output_dir,
                            result_shape,
                            result_dtype,
                            single_file,
                            write_pickle)

        # If requested, perform single-worker dry-run (and quit if desired)
        if dryrun:
            goOn = self.perform_dryrun(setup_interactive)
            if not goOn:
                return

        # Either use existing dask client or start a fresh instance
        self.prepare_client(n_workers=n_workers,
                            partition=partition,
                            mem_per_worker=mem_per_worker,
                            setup_timeout=setup_timeout,
                            setup_interactive=setup_interactive,
                            stop_client=stop_client)

    def initialize(self, func, argv, kwargv, n_calls):
        """
        Parse (provided) inputs: make sure positional and keyword args are
        properly formatted for concurrently calling `func`
        """

        # Allocate slots
        self.func = None
        self.argv = None
        self.kwargv = None
        self.n_calls = None
        self.n_workers = None
        self.acme_func = None
        self.task_ids = None
        self.out_dir = None
        self.collect_results = None
        self.results_container = None
        self.result_shape = None
        self.result_dtype = None
        self.stacking_dim = None
        self.client = None
        self.stop_client = None
        self.has_slurm = None
        self.log = None

        # Ensure `func` is callable
        if not callable(func):
            msg = "{} first input has to be a callable function, not {}"
            raise TypeError(msg.format(self.msgName, str(type(func))))

        # Next, vet `n_calls` which is needed to validate `argv` and `kwargv`
        try:
            acs._scalar_parser(n_calls, varname="n_calls", ntype="int_like", lims=[1, np.inf])
        except Exception as exc:
            raise exc

        # Ensure all elements of `argv` are list-like with lengths `n_calls` or 1
        msg = "{} `argv` has to be a list with list-like elements of length 1 or {}"
        if not isinstance(argv, (list, tuple)):
            raise TypeError(msg.format(self.msgName, n_calls))
        try:
            validArgv = all(len(arg) == n_calls or len(arg) == 1 for arg in argv)
        except TypeError:
            raise TypeError(msg.format(self.msgName, n_calls))
        if not validArgv:
            raise ValueError(msg.format(self.msgName, n_calls))

        # Ensure all values of `kwargv` are list-like with lengths `n_calls` or 1
        msg = "{} `kwargv` has to be a dictionary with list-like elements of length {}"
        try:
            validKwargv = all(len(value) == n_calls or len(value) == 1 for value in kwargv.values())
        except TypeError:
            raise TypeError(msg.format(self.msgName, n_calls))
        if not validKwargv:
            raise ValueError(msg.format(self.msgName, n_calls))

        # Basal sanity checks have passed, keep the provided input signature
        self.func = func
        self.argv = argv
        self.kwargv = kwargv
        self.n_calls = n_calls

        # Define list of taskIDs for distribution across workers
        self.task_ids = list(range(n_calls))

        # Finally, determine if the code is executed on a SLURM-enabled node
        self.has_slurm = acs.is_slurm_node()

    def prepare_output(self,
                       write_worker_results,
                       output_dir,
                       result_shape,
                       result_dtype,
                       single_file,
                       write_pickle):
        """
        If `write_*` is `True` set up directories for saving output HDF5 containers
        (or pickle files). Warn if results are to be collected in memory
        """

        # Basal sanity check for Boolean flags
        if not isinstance(write_worker_results, bool):
            msg = "%s `write_worker_results` has to be `True` or `False`, not %s"
            raise TypeError(msg%(self.msgName, str(write_worker_results)))
        if not isinstance(single_file, bool):
            msg = "%s `single_file` has to be `True` or `False`, not %s"
            raise TypeError(msg%(self.msgName, str(single_file)))
        if not isinstance(write_pickle, bool):
            msg = "%s `write_pickle` has to be `True` or `False`, not %s"
            raise TypeError(msg%(self.msgName, str(write_pickle)))

        # Check compatibility of provided optional args
        if not write_worker_results and write_pickle:
            self.log.warning("Pickling of results only possible if `write_worker_results` is `True`. ")
        if not write_worker_results and output_dir:
            self.log.warning("Output directory specification has no effect if `write_worker_results` is `False`.")
        if write_pickle and result_shape:
            self.log.warning("Pickling of results does not support output array shape specification. ")
        if not write_worker_results and single_file:
            self.log.warning("Generating a single output file only possible if `write_worker_results` is `True`. ")
        if write_pickle and single_file:
            msg = "%s Pickling of results does not support single output file creation. "
            raise ValueError(msg%self.msgName)

        # Check validity of output shape/dtype specifications
        if result_shape is not None:
            if not isinstance(result_shape, (list, tuple)):
                msg = "%s `result_shape` has to be either `None` or tuple, not %s"
                raise TypeError(msg%(self.msgName, str(type(result_shape))))

            if not isinstance(result_dtype, str):
                msg = "%s `result_dtype` has to be a string, not %s"
                raise TypeError(msg%(self.msgName, str(type(result_shape))))

            if sum(spec is None for spec in result_shape) != 1:
                msg = "%s `result_shape` must contain exactly one `None` entry"
                raise ValueError(msg%self.msgName)

            rShape = list(result_shape)
            self.stacking_dim = result_shape.index(None)
            rShape[self.stacking_dim] = self.n_calls
            if write_worker_results:
                self.kwargv["stackingDim"] = [self.stacking_dim]

            if not all(isinstance(spec, numbers.Number) for spec in rShape):
                msg = "%s `result_shape` must only contain numerical values"
                raise ValueError(msg%self.msgName)
            if any(spec < 0 or int(spec) != spec for spec in rShape):
                msg = "%s `result_shape` must only contain non-negative integers"
                raise ValueError(msg%self.msgName)

            self.result_shape = tuple(rShape)
            ShapeSource = list(rShape)
            ShapeSource.pop(self.stacking_dim)
            ShapeSource = tuple(ShapeSource)

            try:
                self.result_dtype = np.dtype(result_dtype)
            except Exception as exc:
                msg = "%s `result_dtype` has to be a valid NumPy datatype specification. "
                msg += "Original error message below:\n%s"
                raise TypeError(msg%(self.msgName, str(exc)))

        # If automatic saving of results is requested, make necessary preparations
        if write_worker_results:

            # Check validity of output dir specification
            if not isinstance(output_dir, (type(None), str)):
                msg = "%s `output_dir` has to be either `None` or str, not %s"
                raise TypeError(msg%(self.msgName, str(type(output_dir))))

            # If provided, standardize output dir spec, otherwise use default locations
            if output_dir is not None:
                outDir = os.path.abspath(os.path.expanduser(output_dir))

            else:
                # On the ESI cluster, save results on HPC mount, otherwise use location of `func`
                if self.has_slurm:
                    outDir = "/cs/home/{usr:s}/".format(usr=getpass.getuser())
                else:
                    outDir = os.path.dirname(os.path.abspath(inspect.getfile(self.func)))
                outDir = os.path.join(outDir, "ACME_{date:s}")
                outDir = outDir.format(date=datetime.datetime.now().strftime('%Y%m%d-%H%M%S-%f'))

            # Unless specifically denied by the user, each worker stores results
            # separately with a common container file pointing to the individual
            # by-worker files residing in a "payload" directory
            self.out_dir = str(outDir)
            if not single_file and not write_pickle:
                payloadName = "{}_payload".format(self.func.__name__)
                outputDir = os.path.join(self.out_dir, payloadName)
            else:
                outputDir = self.out_dir
            try:
                os.makedirs(outputDir)
            except Exception as exc:
                msg = "{} automatic creation of output folder {} failed. Original error message below:\n{}"
                raise OSError(msg.format(self.msgName, outputDir, str(exc)))

            # Re-define or allocate key "taskID" to track concurrent processing results
            self.kwargv["taskID"] = self.task_ids
            self.collect_results = False

            # Set up correct file-extension for output files; in case of HDF5
            # containers, prepare "main" file for collecting/symlinking worker results
            if write_pickle:
                fExt = "pickle"
            else:
                fExt = "h5"
                self.results_container = os.path.join(self.out_dir, "{}.h5".format(self.func.__name__))

            # By default, `results_container` is a collection of links that point to
            # worker-generated HDF5 containers; if `single_file` is `True`, then
            # `results_container` is a "real" container with actual dataset(s)
            if single_file:
                self.kwargv["singleFile"] = [True]
                self.kwargv["outFile"] = [self.results_container]

                # If no output shape provided, prepare groups for storing datasets;
                # otherwise allocate a single dataset w/specified dimension
                if result_shape is None:
                    with h5py.File(self.results_container, "w") as h5f:
                        for i in self.task_ids:
                            h5f.create_group("comp_{}".format(i))
                else:
                    with h5py.File(self.results_container, "w") as h5f:
                        h5f.create_dataset("result_0", shape=self.result_shape, dtype=self.result_dtype)

            else:
                self.kwargv["outFile"] = [os.path.join(outputDir,
                                                       "{}_{}.{}".format(self.func.__name__,
                                                                         taskID,
                                                                         fExt))
                                                       for taskID in self.task_ids]
                if not write_pickle:

                    # If no output shape provided, generate links to external datasets;
                    # otherwise allocate a virtual dataset w/specified dimension
                    if result_shape is None:
                        with h5py.File(self.results_container, "w") as h5f:
                            for i, fname in enumerate(self.kwargv["outFile"]):
                                relPath = os.path.join(payloadName, os.path.basename(fname))
                                h5f["comp_{}".format(i)] = h5py.ExternalLink(relPath, "/")
                    else:

                        # Assemble virtual dataset
                        layout = h5py.VirtualLayout(shape=self.result_shape, dtype=self.result_dtype)
                        idx = [slice(None)] * len(self.result_shape)
                        for i, fname in enumerate(self.kwargv["outFile"]):
                            idx[self.stacking_dim] = i
                            relPath = os.path.join(payloadName, os.path.basename(fname))
                            vsource = h5py.VirtualSource(fname, "result_0", shape=ShapeSource)
                            layout[tuple(idx)] = vsource
                        with h5py.File(self.results_container, "w", libver="latest") as h5f:
                            h5f.create_virtual_dataset("result_0", layout)

            # Include logger name in keywords so that workers can use it
            self.kwargv["logName"] = [self.log.name]

            # Wrap the user-provided func and distribute it across workers
            self.kwargv["userFunc"] = [self.func]
            self.acme_func = self.func_wrapper

        else:

            # If `taskID` is not an explicit kw-arg of `func` and `func` does not
            # accept "anonymous" `**kwargs`, don't save anything but return stuff
            if self.kwargv.get("taskID") is None:
                if not isSpyModule:
                    msg = "`write_worker_results` is `False` and `taskID` is not a keyword argument of {}. " +\
                        "Results will be collected in memory by caller - this might be slow and can lead " +\
                        "to excessive memory consumption. "
                    self.log.warning(msg.format(self.func.__name__))
                self.collect_results = True
            else:
                self.kwargv["taskID"] = self.task_ids
                self.collect_results = False

            # The "raw" user-provided function is used in the computation
            self.acme_func = self.func

    def perform_dryrun(self, setup_interactive):
        """
        Execute user function with one prepared randomly picked args, kwargs combo
        """

        # Let helper randomly pick a single scheduled job and prepare corresponding args + kwargs
        [dryRunIdx], [dryRunArgs], [dryRunKwargs] = self._dryrun_setup(n_runs=1)

        # Create log entry
        msg = "Performing a single dry-run of {fname:s} simulating randomly " +\
            "picked worker #{wrknum:d} with automatically distributed arguments"
        self.log.info(msg.format(fname=self.func.__name__, wrknum=dryRunIdx))

        # Use resident memory size (in MB) to estimate job's memory footprint and measure elapsed time
        mem0 = psutil.Process().memory_info().rss / 1024 ** 2
        tic = time.perf_counter()
        self.acme_func(*dryRunArgs, **dryRunKwargs)
        toc = time.perf_counter()
        mem1 = psutil.Process().memory_info().rss / 1024 ** 2

        # Remove any generated output files
        if self.out_dir is not None:
            os.unlink(self.kwargv["outFile"][dryRunIdx])

        # Compute elapsed time and memory usage
        elapsedTime = toc - tic
        memUsage = mem1 - mem0

        # Prepare info message
        memUnit = "MB"
        if memUsage > 1000:
            memUsage /= 1024
            memUnit = "GB"
        msg = "Dry-run completed. Elapsed time is {runtime:f} seconds, " +\
            "estimated memory consumption was {memused:3.2f} {memunit:s}."
        self.log.info(msg.format(runtime=elapsedTime, memused=memUsage, memunit=memUnit))

        # If the worker setup is supposed to be interactive, ask for confirmation
        # here as well; if execution is terminated, remove auto-generated output directory
        goOn = True
        if setup_interactive:
            msg = "Do you want to continue executing {fname:s} with the provided arguments?"
            if not user_yesno(msg.format(fname=self.func.__name__), default="yes"):
                if self.out_dir is not None:
                    shutil.rmtree(self.out_dir, ignore_errors=True)
                goOn = False

        return goOn

    def _dryrun_setup(self, n_runs=None):
        """
        Pick scheduled job(s) at random and extract corresponding (already prepared!) args + kwargs
        """

        # If not provided, attempt to infer a "sane" default for the number of jobs to pick
        if n_runs is None:
            n_runs = min(self.n_calls, max(5, min(1, int(0.05 * self.n_calls))))

        # Randomly pick `n_runs` jobs and extract positional and keyword args
        dryRunIdx = np.random.choice(self.n_calls, size=n_runs, replace=False)
        dryRunArgs = []
        dryRunKwargs = []
        for idx in dryRunIdx:
            dryRunArgs.append([arg[idx] if len(arg) > 1 else arg[0] for arg in self.argv])
            dryRunKwargs.append([{key:value[idx] if len(value) > 1 else value[0] \
                for key, value in self.kwargv.items()}][0])
        return dryRunIdx, dryRunArgs, dryRunKwargs

    def prepare_client(
        self,
        n_workers="auto",
        partition="auto",
        mem_per_worker="auto",
        setup_timeout=180,
        setup_interactive=True,
        stop_client="auto"):
        """
        Setup or fetch dask distributed processing client. Depending on available
        hardware, either start a local multi-processing client or launch a
        worker cluster via SLURM.

        Also ensure that ad-hoc clients created here are stopped and worker jobs
        are properly released at the end of computation. However, ensure any client
        not created by `prepare_client` is **not** automatically cleaned up.
        """

        # Modify automatic setting of `stop_client` if requested
        msg = "{} `stop_client` has to be 'auto' or Boolean, not {}"
        if isinstance(stop_client, str):
            if stop_client != "auto":
                raise ValueError(msg.format(self.msgName, stop_client))
        elif isinstance(stop_client, bool):
            self.stop_client = stop_client
        else:
            raise TypeError(msg.format(self.msgName, stop_client))

        # Check if a dask client is already running
        try:
            self.client = dd.get_client()
            if stop_client == "auto":
                self.stop_client = False
            self.n_workers = len(self.client.cluster.workers)
            msg = "Attaching to global parallel computing client {}"
            self.log.info(msg.format(str(self.client)))
            return
        except ValueError:
            if stop_client == "auto":
                self.stop_client = True

        # If things are running locally, simply fire up a dask-distributed client,
        # otherwise go through the motions of preparing a full worker cluster
        if not self.has_slurm:
            self.client = local_cluster_setup(interactive=False)

        else:

            # If `partition` is "auto", use `estimate_memuse` to heuristically determine
            # average memory consumption of jobs
            if not isinstance(partition, str):
                msg = "{} `partition` has to be 'auto' or a valid SLURM partition name, not {}"
                raise TypeError(msg.format(self.msgName, str(partition)))
            if partition == "auto":
                if is_esi_node():
                    msg = "Automatic SLURM partition selection is experimental"
                    self.log.warning(msg)
                    mem_per_worker = self.estimate_memuse()
                else:
                    err = "Automatic SLURM partition selection currently only available " +\
                        "on the ESI HPC cluster. "
                    self.log.error(err)

            # If `n_workers` is `"auto`, set `n_workers = n_calls` (default)
            msg = "{} `n_workers` has to be 'auto' or an integer >= 2, not {}"
            if isinstance(n_workers, str):
                if n_workers != "auto":
                    raise ValueError(msg.format(self.msgName, n_workers))
                n_workers = self.n_calls

            # All set, remaining input processing is done by respective `*_cluster_setup` routines
            if is_esi_node():
                self.client = esi_cluster_setup(partition=partition, n_workers=n_workers,
                                                mem_per_worker=mem_per_worker, timeout=setup_timeout,
                                                interactive=setup_interactive, start_client=True)

            # Unknown cluster node, use vanilla config
            else:
                wrng = "Cluster node {} not recognized. Falling back to vanilla " +\
                    "SLURM setup allocating one worker and one core per worker"
                self.log.warning(wrng.format(socket.getfqdn()))
                processes_per_worker = 1
                n_cores = 1
                self.client = slurm_cluster_setup(partition=partition,
                                                  n_cores=n_cores,
                                                  n_workers=n_workers,
                                                  processes_per_worker=processes_per_worker,
                                                  mem_per_worker=mem_per_worker,
                                                  n_workers_startup=1,
                                                  timeout=setup_timeout,
                                                  interactive=setup_interactive,
                                                  interactive_wait=120,
                                                  start_client=True,
                                                  job_extra=[],
                                                  invalid_partitions=[])

            # If startup is aborted by user, get outta here
            if self.client is None:
                msg = "{} Could not start distributed computing client. "
                raise ConnectionAbortedError(msg.format(self.msgName))

        # Set `n_workers` to no. of active workers in the initialized cluster
        self.n_workers = len(self.client.cluster.workers)

        # If single output file saving was chosen, initialize distributed
        # lock for shared writing to container
        if self.kwargv.get("singleFile") is not None:
            dd.lock.Lock(name=os.path.basename(self.results_container))

    def estimate_memuse(self):
        """
        A brute-force guessing approach to determine memory consumption of provided
        workload
        """

        # Let helper randomly pick some jobs and prepare corresponding args + kwargs
        dryRunIdx, dryRunArgs, dryRunKwargs = self._dryrun_setup()

        # Append new dummy keyword to return before any disk-writes happen
        # in case ACME handles results output
        if self.out_dir is not None:
            for k in range(len(dryRunKwargs)):
                dryRunKwargs[k]["memEstRun"] = True

        # Set run-time for each job (in seconds) and allocate arrays for memory
        # consumption estimates (in GB)
        runTime = 30
        memPerSec = np.zeros((runTime,))
        memPerJob = np.zeros((len(dryRunIdx),))

        # # Check if auto-generated output files have to be removed
        # rmOutDir = False
        # if self.out_dir is not None and self.kwargv.get("singleFile") is None:
        #     rmOutDir = True

        # Adequately warn about this heuristic gymnastics...
        msg = "Estimating memory consumption of {fname:s} by running {numwrks:d} " +\
            "random workers for max. {rtime:d} seconds..."
        self.log.info(msg.format(fname=self.func.__name__, numwrks=len(dryRunIdx), rtime=runTime))
        wmsg = "Launching worker #{wrknum:d}"

        for i, idx in enumerate(dryRunIdx):

            # Set up dedicated process to execute user-provided function w/allocated args/kwargs
            proc = multiprocessing.Process(target=self.acme_func,
                                           args=dryRunArgs[i],
                                           kwargs=dryRunKwargs[i])

            # Run user-func for max. `runTime` seconds (or worker finishes),
            # get memory footprint every second
            proc.start()
            with tqdm.tqdm(desc=wmsg.format(wrknum=idx),
                           total=runTime,
                           bar_format=self.tqdmFormat,
                           position=0) as pbar:
                for k in range(runTime):
                    memPerSec[k] = psutil.Process(proc.pid).memory_info().rss / 1024 ** 3
                    time.sleep(1)
                    pbar.update(1)
                    if not proc.is_alive():
                        pbar.n = runTime
                        break
            proc.kill()

            # Compute peak memory consumption across `runTime` seconds
            memPerJob[i] = memPerSec.max()

        # Compute aggregate average memory consumption across all runs
        memUsage = memPerJob.mean()

        # Communicate results
        msg = "Estimated memory consumption across {numwrks:d} runs is {memuse:3.2f} GB "
        self.log.info(msg.format(numwrks=len(dryRunIdx), memuse=memUsage))

        return "estimate_memuse:" + str(max(1, int(np.ceil(memUsage))))

    def compute(self, debug=False):
        """
        Perform the actual parallel execution of `func`

        If `debug` is `True`, use a single-threaded dask scheduler that does
        not actually process anything concurrently but uses the dask framework
        in a sequential setup.
        """

        # If `prepare_client` has not been called yet, don't attempt to compute anything
        if self.client is None:
            return

        # Ensure `debug` is a simple Boolean flag
        if not isinstance(debug, bool):
            msg = "{} `debug` has to be `True` or `False`, not {}"
            raise TypeError(msg.format(self.msgName, str(debug)))

        # Deduce result output information
        write_worker_results = self.acme_func == self.func_wrapper
        single_file = False
        if write_worker_results:
            write_pickle = self.results_container is None
            if not write_pickle and self.kwargv.get("singleFile") is not None:
                single_file = True
        else:
            write_pickle = False

        # Check if the underlying parallel computing cluster hosts actually usable workers
        if len([w["memory_limit"] for w in self.client.cluster.scheduler_info["workers"].values()]) == 0:
            msg = "{} no active workers found in distributed computing cluster {} " +\
                "Consider running \n" +\
                "\timport dask.distributed as dd; dd.get_client().restart()\n" +\
                "If this fails to make workers come online, please use\n" +\
                "\timport acme; acme.cluster_cleanup()\n" +\
                "to shut down any defunct distributed computing clients"
            raise RuntimeError(msg.format(self.msgName, self.client))

        # Dask does not correctly forward the `sys.path` from the parent process
        # to its workers. Fix this.
        def init_acme(dask_worker, syspath):
            sys.path = list(syspath)
        self.client.register_worker_callbacks(setup=functools.partial(init_acme, syspath=sys.path))

        # Format positional arguments for worker-distribution: broadcast all
        # inputs that are used by all workers and create a list of references
        # to this (single!) future on the cluster for submission
        for ak, arg in enumerate(self.argv):
            if len(arg) == 1:
                ftArg = self.client.scatter(arg, broadcast=True)
                if isinstance(ftArg, collections.abc.Sized):
                    ftArg = ftArg[0]
                self.argv[ak] = [ftArg] * self.n_calls

        # Same as above but for keyword-arguments
        for name, value in self.kwargv.items():
            if len(value) == 1:
                ftVal = self.client.scatter(value, broadcast=True)[0]
                self.kwargv[name] = [ftVal] * self.n_calls

        # Re-format keyword arguments to be usable with single-to-many arg submission.
        # Idea: with `self.n_calls = 3` and ``self.kwargv = {'a': [5, 5, 5], 'b': [6, 6, 6]}``
        # then ``kwargList = [{'a': 5, 'b': 6}, {'a': 5, 'b': 6}, {'a': 5, 'b': 6}]``
        kwargList = []
        kwargKeys = self.kwargv.keys()
        kwargVals = list(self.kwargv.values())
        for nc in range(self.n_calls):
            kwDict = {}
            for kc, key in enumerate(kwargKeys):
                kwDict[key] = kwargVals[kc][nc]
            kwargList.append(kwDict)

        # In case a debugging run is performed, use the single-threaded scheduler and return
        if debug:
            with dask.config.set(scheduler='single-threaded'):
                values = self.client.gather([self.client.submit(self.acme_func, *args, **kwargs) \
                    for args, kwargs in zip(zip(*self.argv), kwargList)])
                return values

        # Depending on the used dask cluster object, point to respective log info
        if isinstance(self.client.cluster, dj.SLURMCluster):
            logFiles = self.client.cluster.job_header.split("--output=")[1].replace("%j", "{}")
            logDir = os.path.split(logFiles)[0]
        else:
            logFiles = []
            logDir = os.path.dirname(self.client.cluster.dashboard_link) + "/info/main/workers.html"
        msg = "Preparing {} parallel calls of `{}` using {} workers"
        self.log.info(msg.format(self.n_calls, self.func.__name__, self.n_workers))
        msg = "Log information available at {}"
        self.log.info(msg.format(logDir))

        # Submit `self.n_calls` function calls to the cluster
        futures = [self.client.submit(self.acme_func, *args, **kwargs) \
            for args, kwargs in zip(zip(*self.argv), kwargList)]

        # Set up progress bar: the while loop ensures all futures are executed
        totalTasks = len(futures)
        pbar = tqdm.tqdm(total=totalTasks, bar_format=self.tqdmFormat, position=0, leave=True)
        cnt = 0
        while any(f.status == "pending" for f in futures):
            time.sleep(self.sleepTime)
            new = max(0, sum([f.status == "finished" for f in futures]) - cnt)
            cnt += new
            pbar.update(new)
        pbar.close()

        # Avoid race condition: give futures time to perform switch from 'pending'
        # to 'finished' so that `finishedTasks` is computed correctly
        time.sleep(self.sleepTime)

        # If number of 'finished' tasks is less than expected, go into
        # problem analysis mode: all futures that erred hav an `.exception`
        # method which can be used to track down the worker it was executed by
        # Once we know the worker, we can point to the right log file. If
        # futures were cancelled (by the user or the SLURM controller),
        # `.exception` is `None` and we can't reliably track down the
        # respective executing worker
        finishedTasks = sum([f.status == "finished" for f in futures])
        if finishedTasks < totalTasks:
            schedulerLog = list(self.client.cluster.get_logs(cluster=False, scheduler=True, workers=False).values())[0]
            erredFutures = [f for f in futures if f.status == "error"]
            msg = "{} Parallel computation failed: {}/{} tasks failed or stalled.\n"
            msg = msg.format(self.msgName, totalTasks - finishedTasks, totalTasks)
            msg += "Concurrent computing scheduler log below: \n\n"
            msg += schedulerLog + "\n"

            # If we're working w/`SLURMCluster`, perform the Herculean task of
            # tracking down which dask worker was executed by which SLURM job...
            if self.client.cluster.__class__.__name__ == "SLURMCluster":
                try:
                    erredJobs = [f.exception().last_worker.identity()["id"] for f in erredFutures]
                except AttributeError:
                    erredJobs = []
                erredJobs = list(set(erredJobs))
                validIDs = [job for job in erredJobs if job in self.client.cluster.workers.keys()]
                erredJobIDs = [self.client.cluster.workers[job].job_id for job in validIDs]
                errFiles = glob.glob(logDir + os.sep + "*.err")
                if len(erredFutures) > 0 or len(errFiles) > 0:
                    msg += "Please consult the following SLURM log files for details:\n"
                    if len(erredJobIDs) > 0:
                        msg += "".join(logFiles.format(id) + "\n" for id in erredJobIDs)
                    else:
                        msg += "".join(logDir)
                    msg += "".join(errfile + "\n" for errfile in errFiles)
                else:
                    msg += "Please check SLURM logs in {}".format(logDir)

            # In case of a `LocalCluster`, syphon worker logs
            else:
                msg += "\nParallel worker logs below: \n"
                workerLogs = self.client.get_worker_logs().values()
                for wLog in workerLogs:
                    if "Failed" in wLog:
                        msg += wLog

            # Finally, raise an error and get outta here
            raise RuntimeError(msg)

        # If wanted (not recommended) collect computed results in local memory
        # The return `values` is either
        # `None` : if neither in-memory results collection or auto-writing was requested
        # list of file-names: if `write_worker_results` is `True`
        # list of objects: if in-memory results collection was requested
        if self.collect_results:
            if not isSpyModule:
                self.log.info("Gathering results in local memory")
            collected = self.client.gather(futures)
            if self.result_shape is not None:
                values = []
                arrVal = np.empty(shape=self.result_shape, dtype=self.result_dtype)
                idx = [slice(None)] * len(self.result_shape)
                for i, res in enumerate(collected):
                    if not isinstance(res, (list, tuple)):
                        res = [res]
                    idx[self.stacking_dim] = i
                    arrVal[tuple(idx)] = res[0]
                    for r in res[1:]:
                        values.append(r)
                values.insert(0, arrVal)
            else:
                values = collected
        else:
            values = None

        # Prepare final output message
        finalMsg = "{}Finished parallel computation. "
        successMsg = "SUCCESS! "

        # If automatic results writing was requested, perform some housekeeping
        if write_worker_results:
            if write_pickle:
                values = list(self.kwargv["outFile"])
                finalMsg += "Results have been saved to {}".format(self.out_dir)
            else:
                if single_file:
                    finalMsg += "Results have been saved to {}".format(self.results_container)
                    if values is None:
                        values = [self.results_container]
                else:
                    picklesFound = False
                    values = []
                    for fname in self.kwargv["outFile"]:
                        pklName = fname.rstrip(".h5") + ".pickle"
                        if os.path.isfile(fname):
                            values.append(fname)
                        elif os.path.isfile(pklName):
                            values.append(pklName)
                            picklesFound = True
                        else:
                            values.append("Missing {}".format(fname.rstrip(".h5")))
                    payloadDir = os.path.dirname(values[0])

                    # If pickles are found, remove global `results_container` as it
                    # would contain invalid file-links and move compute results out
                    # of payload dir
                    if picklesFound:
                        os.unlink(self.results_container)
                        wrng = "Some compute runs could not be saved as HDF5, " +\
                            "collection container %s has been removed as it would " +\
                                "comprise invalid file-links"
                        self.log.warning(wrng, self.results_container)
                        self.results_container = None

                        # Move files out of payload dir and update return `values`
                        target = os.path.abspath(os.path.join(payloadDir, os.pardir))
                        for i, fname in enumerate(values):
                            shutil.move(fname, target)
                            self.kwargv["outFile"][i] = os.path.join(target, os.path.basename(fname))
                        values = list(self.kwargv["outFile"])
                        shutil.rmtree(payloadDir)
                        successMsg = ""
                        finalMsg += "Results have been saved to {}".format(target)

                    # All good, no pickle gymnastics was needed
                    else:

                        # In case of multiple return values present in by-worker
                        # containers but missing in collection container (happens
                        # if `result_shape` is not `None` and data-sets have to
                        # be pre-allocated), create "symlinks" to corresponding
                        # missing returns
                        if self.stacking_dim is not None:
                            with h5py.File(self.results_container, "r") as h5r:
                                with h5py.File(values[0], "r") as h5Tmp:
                                    missingReturns = set(h5Tmp.keys()).difference(h5r.keys())
                            if len(missingReturns) > 0:
                                with h5py.File(self.results_container, "a") as h5r:
                                    for retVal in missingReturns:
                                        for i, fname in enumerate(values):
                                            relPath = os.path.join(os.path.basename(payloadDir), os.path.basename(fname))
                                            h5r["comp_{}/{}".format(i, retVal)] = h5py.ExternalLink(relPath, retVal)

                        msg = "Results have been saved to {} with links to data payload located in {}"
                        finalMsg += msg.format(self.results_container, payloadDir)

        # Print final triumphant output message and get out
        self.log.info(finalMsg.format(successMsg))

        # Either return collected by-worker results or the filepaths of results
        return values

    def cleanup(self):
        """
        Shut down any ad-hoc distributed computing clients created by `prepare_client`
        """

        # If `prepare_client` has not been launched yet, just get outta here
        if not hasattr(self, "client"):
            return
        if self.stop_client and self.client is not None:
            cluster_cleanup(self.client)
            self.client = None

    @staticmethod
    def func_wrapper(*args, **kwargs):
        """
        If the output of `func` is saved to disk, wrap `func` with this static
        method to take care of filling up HDF5/pickle files

        If writing to HDF5 fails, use an "emergency-pickling" mechanism to try
        to save the output of `func` using pickle instead
        """

        # Extract everything from `kwargs` appended by `ACMEdaemon`
        func = kwargs.pop("userFunc")
        taskID = kwargs.pop("taskID")
        fname = kwargs.pop("outFile")
        logName = kwargs.pop("logName")
        singleFile = kwargs.pop("singleFile", False)
        stackingDim = kwargs.pop("stackingDim", None)
        memEstRun = kwargs.pop("memEstRun", False)
        log = logging.getLogger(logName)

        # Call user-provided function
        result = func(*args, **kwargs)

        # For memory estimation runs, don't start saving stuff
        if memEstRun:
            return

        # Save results: either (try to) use HDF5 or pickle stuff
        if fname.endswith(".h5"):

            grpName = ""
            if singleFile:
                lock = dd.lock.Lock(name=os.path.basename(fname))
                lock.acquire()
                grpName = "comp_{}/".format(taskID)
                errName = pname = fname.rstrip(".h5") + ".failed"
                err = "Could not write to %s. File potentially corrupted. "
                if os.path.isfile(errName):
                    log.error(err, fname)
                    lock.release()
                    raise IOError(err%fname)

            if not isinstance(result, (list, tuple)):
                result = [result]

            try:

                with h5py.File(fname, "a") as h5f:
                    if stackingDim is None:
                        if not all(isinstance(value, (numbers.Number, str)) for value in result):
                            for rk, res in enumerate(result):
                                h5f.create_dataset(grpName + "result_{}".format(rk), data=res)
                        else:
                            h5f.create_dataset(grpName + "result_0", data=result)
                    else:
                        if singleFile:
                            idx = [slice(None)] * len(h5f["result_0"].shape)
                            idx[stackingDim] = taskID
                            h5f["result_0"][tuple(idx)] = result[0]
                            for rk, res in enumerate(result[1:]):
                                h5f.create_dataset(grpName + "result_{}".format(rk + 1), data=res)
                        else:
                            for rk, res in enumerate(result):
                                h5f.create_dataset(grpName + "result_{}".format(rk), data=res)

                if singleFile:
                    lock.release()

            except TypeError as exc:

                if ("has no native HDF5 equivalent" in str(exc) \
                    or "One of data, shape or dtype must be specified" in str(exc)) \
                        and not singleFile:
                    try:
                        os.unlink(fname)
                        pname = fname.rstrip(".h5") + ".pickle"
                        with open(os.path.join(pname), "wb") as pkf:
                            pickle.dump(result, pkf)
                        msg = "Could not write %s results have been pickled instead: %s. Return values are most likely " +\
                            "not suitable for storage in HDF5 containers. Original error message: %s"
                        log.warning(msg, fname, pname, str(exc))
                    except pickle.PicklingError as pexc:
                        err = "Unable to write %s, successive attempts to pickle results failed too: %s"
                        log.error(err, fname, str(pexc))
                else:
                    if singleFile:
                        err = "Could not write to %s. File potentially corrupted. Original error message: %s"
                        lock.release()
                    else:
                        err = "Could not access %s. Original error message: %s"
                    log.error(err, fname, str(exc))
                    raise exc

            except Exception as exc:

                if singleFile:
                    lock.release()
                raise exc

        else:

            try:
                with open(os.path.join(fname), "wb") as pkf:
                    pickle.dump(result, pkf)
            except pickle.PicklingError as pexc:
                err = "Could not pickle results to file %s. Original error message: %s"
                log.error(err, fname, str(pexc))
                raise pexc
